# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.v2.components.experimental.component_spec."""

import unittest
from unittest.mock import patch, mock_open
from kfp.v2.components.experimental.component_spec import ComponentSpec
from kfp.v2.components.experimental import component_spec
import textwrap


class ComponentSpecTest(unittest.TestCase):

    @unittest.skip("Placeholder check is not completed. ")
    def test_component_spec_with_placeholder_referencing_nonexisting_input_output(
            self):
        with self.assertRaisesRegex(
                ValueError,
                'Argument "InputValuePlaceholder\(name=\'input000\'\)" '
                'references non-existing input.'):
            component_spec.ComponentSpec(
                name='component_1',
                implementation=component_spec.ContainerSpec(
                    image='alpine',
                    commands=[
                        'sh',
                        '-c',
                        'set -ex\necho "$0" > "$1"',
                        component_spec.InputValuePlaceholder(name='input000'),
                        component_spec.OutputPathPlaceholder(name='output1'),
                    ],
                ),
                input_specs=[
                    component_spec.InputSpec(name='input1', type='String'),
                ],
                output_specs=[
                    component_spec.OutputSpec(name='output1', type='String'),
                ],
            )

        with self.assertRaisesRegex(
                ValueError,
                'Argument "OutputPathPlaceholder\(name=\'output000\'\)" '
                'references non-existing output.'):
            component_spec.ComponentSpec(
                name='component_1',
                implementation=component_spec.ContainerSpec(
                    image='alpine',
                    commands=[
                        'sh',
                        '-c',
                        'set -ex\necho "$0" > "$1"',
                        component_spec.InputValuePlaceholder(name='input1'),
                        component_spec.OutputPathPlaceholder(name='output000'),
                    ],
                ),
                input_specs=[
                    component_spec.InputSpec(name='input1', type='String'),
                ],
                output_specs=[
                    component_spec.OutputSpec(name='output1', type='String'),
                ],
            )

    def test_component_spec_save_to_component_yaml(self):
        open_mock = mock_open()
        expected_yaml = textwrap.dedent("""\
        implementation:
          commands:
          - sh
          - -c
          - 'set -ex

            echo "$0" > "$1"'
          - name: input1
          - name: output1
          image: alpine
        inputs:
          input1:
            type: String
        name: component_1
        outputs:
          output1:
            type: String
        schema_version: v2
        """)

        with patch("builtins.open", open_mock, create=True):
            component_spec.ComponentSpec(
                name='component_1',
                implementation=component_spec.ContainerSpec(
                    image='alpine',
                    commands=[
                        'sh',
                        '-c',
                        'set -ex\necho "$0" > "$1"',
                        component_spec.InputValuePlaceholder(name='input1'),
                        component_spec.OutputPathPlaceholder(name='output1'),
                    ],
                ),
                inputs={
                    'input1': component_spec.InputSpec(type='String')
                },
                outputs={
                    'output1': component_spec.OutputSpec(type='String')
                },
                schema_version=component_spec.SchemaVersion.V2,
            ).save_to_component_yaml('test_save_file.txt')

        open_mock.assert_called_with('test_save_file.txt', 'a')
        open_mock.return_value.write.assert_called_once_with(expected_yaml)

    def test_component_spec_load_from_v2_component_yaml(self):
        component_yaml_v2 = textwrap.dedent("""\
        implementation:
          commands:
          - sh
          - -c
          - 'set -ex

            echo "$0" > "$1"'
          - name: input1
          - name: output1
          image: alpine
        inputs:
          input1:
            type: String
        name: component_1
        outputs:
          output1:
            type: String
        schema_version: v2
        """)

        generated_spec = component_spec.ComponentSpec.load_from_component_yaml(
            component_yaml_v2)

        expected_spec = component_spec.ComponentSpec(
            name='component_1',
            implementation=component_spec.ContainerSpec(
                image='alpine',
                commands=[
                    'sh',
                    '-c',
                    'set -ex\necho "$0" > "$1"',
                    component_spec.InputValuePlaceholder(name='input1'),
                    component_spec.OutputPathPlaceholder(name='output1'),
                ],
            ),
            inputs={'input1': component_spec.InputSpec(type='String')},
            outputs={'output1': component_spec.OutputSpec(type='String')},
            schema_version=component_spec.SchemaVersion.V2)
        self.assertEqual(generated_spec, expected_spec)

    def test_component_spec_load_from_v1_component_yaml(self):
        component_yaml_v1 = textwrap.dedent("""\
        name: Component with 2 inputs and 2 outputs
        inputs:
        - {name: Input parameter}
        - {name: Input artifact}
        outputs:
        - {name: Output 1}
        - {name: Output 2}
        implementation:
          container:
            image: busybox
            command: [sh, -c, '
                mkdir -p $(dirname "$2")
                mkdir -p $(dirname "$3")
                echo "$0" > "$2"
                cp "$1" "$3"
                '
            ]
            args:
            - {inputValue: Input parameter}
            - {inputPath: Input artifact}
            - {outputPath: Output 1}
            - {outputPath: Output 2}
        """)

        generated_spec = component_spec.ComponentSpec.load_from_component_yaml(
            component_yaml_v1)

        expected_spec = component_spec.ComponentSpec(
            name='Component with 2 inputs and 2 outputs',
            implementation=component_spec.ContainerSpec(
                image='busybox',
                commands=[
                    'sh',
                    '-c',
                    (' mkdir -p $(dirname "$2") mkdir -p $(dirname "$3") '
                     'echo "$0" > "$2" cp "$1" "$3" '),
                ],
                arguments=[
                    component_spec.InputValuePlaceholder(
                        name='Input parameter'),
                    component_spec.InputPathPlaceholder(name='Input artifact'),
                    component_spec.OutputPathPlaceholder(name='Output 1'),
                    component_spec.OutputPathPlaceholder(name='Output 2'),
                ],
                env={},
            ),
            inputs={
                'Input parameter': component_spec.InputSpec(type='Artifact'),
                'Input artifact': component_spec.InputSpec(type='Artifact')
            },
            outputs={
                'Output 1': component_spec.OutputSpec(type='String'),
                'Output 2': component_spec.OutputSpec(type='String'),
            },
            schema_version=component_spec.SchemaVersion.V1)

        self.assertEqual(generated_spec, expected_spec)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for KFP components."""

import abc

from kfp.v2.components.experimental import component_spec as cspec
from kfp.v2.components.experimental import pipeline_task


class BaseComponent(metaclass=abc.ABCMeta):
    """Base class for a component.

    Attributes:
      name: The name of the component.
      component_spec: The component definition.
    """

    def __init__(self, component_spec: cspec.ComponentSpec):
        """Init function for BaseComponent.

        Args:
          component_spec: The component definition.
        """
        self.component_spec = component_spec
        self.name = component_spec.name

        self._component_inputs = set(self.component_spec.inputs.keys())

    def __call__(self, *args, **kwargs) -> pipeline_task.PipelineTask:
        """Creates a PipelineTask object.

        The arguments are generated on the fly based on component input
        definitions.
        """
        task_inputs = {}

        if len(args) > 0:
            raise TypeError(
                'Components must be instantiated using keyword arguments. Positional '
                f'parameters are not allowed (found {len(args)} such parameters for '
                f'component "{self.name}").')

        for k, v in kwargs.items():
            if k not in self._component_inputs:
                raise TypeError(
                    f'{self.name}() got an unexpected keyword argument "{k}".')

            if k in task_inputs:
                raise TypeError(
                    f'{self.name}() got multiple values for argument "{k}".')
            task_inputs[k] = v

        # Fill in default value if there was no user provided value
        for name, input_spec in self.component_spec.inputs.items():
            if input_spec.default is not None and name not in task_inputs:
                task_inputs[name] = input_spec.default

        missing_arguments = [
            name for name in self.component_spec.inputs
            if name not in task_inputs
        ]
        if missing_arguments:
            argument_or_arguments = 'argument' if len(
                missing_arguments) == 1 else 'arguments'
            arguments = ','.join(missing_arguments)

            raise TypeError(
                f'{self.name}() missing {len(missing_arguments)} required positional '
                f'{argument_or_arguments}: {arguments}.')

        return pipeline_task.create_pipeline_task(
            component_spec=self.component_spec,
            arguments=task_inputs,
        )

    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        """Executes the component given the required inputs.

        Subclasses of BaseComponent must override this abstract method
        in order to be instantiated. For Python function-based
        component, the implementation of this method could be calling
        the function. For "Bring your own container" component, the
        implementation of this method could be `docker run`.
        """
        pass



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""KFP SDK compiler CLI tool."""

import argparse
import json
import os
import sys
from typing import Any, Callable, List, Mapping, Optional

import kfp.dsl as dsl
import kfp.v2.compiler as compiler


def parse_arguments() -> argparse.Namespace:
    """Parse command line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--py',
        type=str,
        required=True,
        help='local absolute path to a py file.')
    parser.add_argument(
        '--function',
        type=str,
        help='The name of the function to compile if there are multiple.')
    parser.add_argument(
        '--pipeline-parameters',
        type=json.loads,
        help='The pipeline parameters in JSON dict format.')
    parser.add_argument(
        '--namespace', type=str, help='The namespace for the pipeline function')
    parser.add_argument(
        '--output',
        type=str,
        required=True,
        help='local path to the output PipelineJob json file.')
    parser.add_argument(
        '--disable-type-check',
        action='store_true',
        help='disable the type check, default is enabled.')

    args = parser.parse_args()
    return args


def _compile_pipeline_function(pipeline_funcs: List[Callable],
                               function_name: Optional[str],
                               pipeline_parameters: Optional[Mapping[str, Any]],
                               package_path: str, type_check: bool) -> None:
    """Compiles a pipeline function.

    Args:
      pipeline_funcs: A list of pipeline_functions.
      function_name: The name of the pipeline function to compile if there were
        multiple.
      pipeline_parameters: The pipeline parameters as a dict of {name: value}.
      package_path: The output path of the compiled result.
      type_check: Whether to enable the type checking.
    """
    if len(pipeline_funcs) == 0:
        raise ValueError(
            'A function with @dsl.pipeline decorator is required in the py file.'
        )

    if len(pipeline_funcs) > 1 and not function_name:
        func_names = [x.__name__ for x in pipeline_funcs]
        raise ValueError(
            'There are multiple pipelines: %s. Please specify --function.' %
            func_names)

    if function_name:
        pipeline_func = next(
            (x for x in pipeline_funcs if x.__name__ == function_name), None)
        if not pipeline_func:
            raise ValueError('The function "%s" does not exist. '
                             'Did you forget @dsl.pipeline decoration?' %
                             function_name)
    else:
        pipeline_func = pipeline_funcs[0]

    compiler.Compiler().compile(
        pipeline_func=pipeline_func,
        pipeline_parameters=pipeline_parameters,
        package_path=package_path,
        type_check=type_check)


class PipelineCollectorContext():

    def __enter__(self):
        pipeline_funcs = []

        def add_pipeline(func: Callable) -> Callable:
            pipeline_funcs.append(func)
            return func

        self.old_handler = dsl._pipeline._pipeline_decorator_handler
        dsl._pipeline._pipeline_decorator_handler = add_pipeline
        return pipeline_funcs

    def __exit__(self, *args):
        dsl._pipeline._pipeline_decorator_handler = self.old_handler


def compile_pyfile(pyfile: str, function_name: Optional[str],
                   pipeline_parameters: Optional[Mapping[str, Any]],
                   package_path: str, type_check: bool) -> None:
    """Compiles a pipeline written in a .py file.

    Args:
      pyfile: The path to the .py file that contains the pipeline definition.
      function_name: The name of the pipeline function.
      pipeline_parameters: The pipeline parameters as a dict of {name: value}.
      package_path: The output path of the compiled result.
      type_check: Whether to enable the type checking.
    """
    sys.path.insert(0, os.path.dirname(pyfile))
    try:
        filename = os.path.basename(pyfile)
        with PipelineCollectorContext() as pipeline_funcs:
            __import__(os.path.splitext(filename)[0])
        _compile_pipeline_function(
            pipeline_funcs=pipeline_funcs,
            function_name=function_name,
            pipeline_parameters=pipeline_parameters,
            package_path=package_path,
            type_check=type_check)
    finally:
        del sys.path[0]


def main():
    args = parse_arguments()
    if args.py is None:
        raise ValueError('The --py option must be specified.')
    compile_pyfile(
        pyfile=args.py,
        function_name=args.function,
        pipeline_parameters=args.pipeline_parameters,
        package_path=args.output,
        type_check=not args.disable_type_check,
    )



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.v2.google.client.runtime_config_builder."""

import unittest

import frozendict
from kfp.v2.google.client import runtime_config_builder


class RuntimeConfigBuilderTest(unittest.TestCase):

    SAMPLE_JOB_SPEC = frozendict.frozendict({
        'pipelineSpec': {
            'root': {
                'inputDefinitions': {
                    'parameters': {
                        'string_param': {
                            'type': 'STRING'
                        },
                        'int_param': {
                            'type': 'INT'
                        },
                        'float_param': {
                            'type': 'DOUBLE'
                        },
                        'new_param': {
                            'type': 'STRING'
                        },
                        'bool_param': {
                            'type': 'STRING'
                        },
                        'dict_param': {
                            'type': 'STRING'
                        },
                        'list_param': {
                            'type': 'STRING'
                        },
                    }
                }
            }
        },
        'runtimeConfig': {
            'gcsOutputDirectory': 'path/to/my/root',
            'parameters': {
                'string_param': {
                    'stringValue': 'test-string'
                },
                'int_param': {
                    'intValue': 42
                },
                'float_param': {
                    'doubleValue': 3.14
                },
            }
        }
    })

    def testBuildRuntimeConfigFromIndividualValues(self):
        my_builder = runtime_config_builder.RuntimeConfigBuilder(
            pipeline_root='path/to/my/root',
            parameter_types={
                'string_param': 'STRING',
                'int_param': 'INT',
                'float_param': 'DOUBLE'
            },
            parameter_values={
                'string_param': 'test-string',
                'int_param': 42,
                'float_param': 3.14
            })
        actual_runtime_config = my_builder.build()

        expected_runtime_config = self.SAMPLE_JOB_SPEC['runtimeConfig']
        self.assertEqual(expected_runtime_config, actual_runtime_config)

    def testBuildRuntimeConfigFromJobSpecJson(self):
        my_builder = (
            runtime_config_builder.RuntimeConfigBuilder.from_job_spec_json(
                self.SAMPLE_JOB_SPEC))
        actual_runtime_config = my_builder.build()

        expected_runtime_config = self.SAMPLE_JOB_SPEC['runtimeConfig']
        self.assertEqual(expected_runtime_config, actual_runtime_config)

    def testBuildRuntimeConfigWithNoopUpdates(self):
        my_builder = (
            runtime_config_builder.RuntimeConfigBuilder.from_job_spec_json(
                self.SAMPLE_JOB_SPEC))
        my_builder.update_pipeline_root(None)
        my_builder.update_runtime_parameters(None)
        actual_runtime_config = my_builder.build()

        expected_runtime_config = self.SAMPLE_JOB_SPEC['runtimeConfig']
        self.assertEqual(expected_runtime_config, actual_runtime_config)

    def testBuildRuntimeConfigWithMergeUpdates(self):
        my_builder = (
            runtime_config_builder.RuntimeConfigBuilder.from_job_spec_json(
                self.SAMPLE_JOB_SPEC))
        my_builder.update_pipeline_root('path/to/my/new/root')
        my_builder.update_runtime_parameters({
            'int_param': 888,
            'new_param': 'new-string',
            'dict_param': {
                'a': 1
            },
            'list_param': [1, 2, 3],
            'bool_param': True,
        })
        actual_runtime_config = my_builder.build()

        expected_runtime_config = {
            'gcsOutputDirectory': 'path/to/my/new/root',
            'parameters': {
                'string_param': {
                    'stringValue': 'test-string'
                },
                'int_param': {
                    'intValue': 888
                },
                'float_param': {
                    'doubleValue': 3.14
                },
                'new_param': {
                    'stringValue': 'new-string'
                },
                'dict_param': {
                    'stringValue': '{"a": 1}'
                },
                'list_param': {
                    'stringValue': '[1, 2, 3]'
                },
                'bool_param': {
                    'stringValue': 'true'
                },
            }
        }
        self.assertEqual(expected_runtime_config, actual_runtime_config)

    def testBuildRuntimeConfigParameterNotFound(self):
        my_builder = (
            runtime_config_builder.RuntimeConfigBuilder.from_job_spec_json(
                self.SAMPLE_JOB_SPEC))
        my_builder.update_pipeline_root('path/to/my/new/root')
        my_builder.update_runtime_parameters({'no_such_param': 'new-string'})
        with self.assertRaisesRegex(
                ValueError,
                'The pipeline parameter no_such_param is not found'):
            my_builder.build()

    def testParseRuntimeParameters(self):
        expected_runtime_parameters = {
            'string_param': 'test-string',
            'int_param': 42,
            'float_param': 3.14,
        }
        actual_runtime_parameters = (
            runtime_config_builder._parse_runtime_parameters(
                self.SAMPLE_JOB_SPEC['runtimeConfig']))
        self.assertEqual(expected_runtime_parameters, actual_runtime_parameters)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for AIPlatformPipelines API client utils."""

import json
from typing import Any, Dict

from google.cloud import storage


def load_json(path: str) -> Dict[str, Any]:
    """Loads data from a JSON document.

    Args:
      path: The path of the JSON document. It can be a local path or a GS URI.

    Returns:
      A deserialized Dict object representing the JSON document.
    """

    if path.startswith('gs://'):
        return _load_json_from_gs_uri(path)
    else:
        return _load_json_from_local_file(path)


def _load_json_from_gs_uri(uri: str) -> Dict[str, Any]:
    """Loads data from a JSON document referenced by a GS URI.

    Args:
      uri: The GCS URI of the JSON document.

    Returns:
      A deserialized Dict object representing the JSON document.

    Raises:
      google.cloud.exceptions.NotFound: If the blob is not found.
      json.decoder.JSONDecodeError: On JSON parsing problems.
      ValueError: If uri is not a valid gs URI.
    """
    storage_client = storage.Client()
    blob = storage.Blob.from_string(uri, storage_client)
    return json.loads(blob.download_as_bytes())


def _load_json_from_local_file(file_path: str) -> Dict[str, Any]:
    """Loads data from a JSON local file.

    Args:
      file_path: The local file path of the JSON document.

    Returns:
      A deserialized Dict object representing the JSON document.

    Raises:
      json.decoder.JSONDecodeError: On JSON parsing problems.
    """
    with open(file_path) as f:
        return json.load(f)



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.v2.components.utils."""

import unittest

from absl.testing import parameterized
from kfp.v2.components import utils


class UtilsTest(parameterized.TestCase):

    @parameterized.parameters(
        {
            'original': 'name',
            'expected': 'name',
        },
        {
            'original': ' Some name',
            'expected': 'some-name',
        },
        {
            'original': 'name_123*',
            'expected': 'name-123',
        },
    )
    def test_maybe_rename_for_k8s(self, original, expected):
        self.assertEqual(utils.maybe_rename_for_k8s(original), expected)


if __name__ == '__main__':
    unittest.main()



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import parameterized

import sys
import unittest
from typing import Any, Dict, List

from kfp.components import structures
from kfp.pipeline_spec import pipeline_spec_pb2 as pb
from kfp.v2.components.types import artifact_types, type_utils

_PARAMETER_TYPES = [
    'String',
    'str',
    'Integer',
    'int',
    'Float',
    'Double',
    'bool',
    'Boolean',
    'Dict',
    'List',
    'JsonObject',
    'JsonArray',
    {
        'JsonObject': {
            'data_type': 'proto:tfx.components.trainer.TrainArgs'
        }
    },
]
_KNOWN_ARTIFACT_TYPES = ['Model', 'Dataset', 'Schema', 'Metrics']
_UNKNOWN_ARTIFACT_TYPES = [None, 'Arbtrary Model', 'dummy']


class _ArbitraryClass:
    pass


class TypeUtilsTest(parameterized.TestCase):

    def test_is_parameter_type(self):
        for type_name in _PARAMETER_TYPES:
            self.assertTrue(type_utils.is_parameter_type(type_name))
        for type_name in _KNOWN_ARTIFACT_TYPES + _UNKNOWN_ARTIFACT_TYPES:
            self.assertFalse(type_utils.is_parameter_type(type_name))

    @parameterized.parameters(
        {
            'artifact_class_or_type_name':
                'Model',
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.Model', schema_version='0.0.1')
        },
        {
            'artifact_class_or_type_name':
                artifact_types.Model,
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.Model', schema_version='0.0.1')
        },
        {
            'artifact_class_or_type_name':
                'Dataset',
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.Dataset', schema_version='0.0.1')
        },
        {
            'artifact_class_or_type_name':
                artifact_types.Dataset,
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.Dataset', schema_version='0.0.1')
        },
        {
            'artifact_class_or_type_name':
                'Metrics',
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.Metrics', schema_version='0.0.1')
        },
        {
            'artifact_class_or_type_name':
                artifact_types.Metrics,
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.Metrics', schema_version='0.0.1')
        },
        {
            'artifact_class_or_type_name':
                'ClassificationMetrics',
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.ClassificationMetrics',
                    schema_version='0.0.1')
        },
        {
            'artifact_class_or_type_name':
                artifact_types.ClassificationMetrics,
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.ClassificationMetrics',
                    schema_version='0.0.1')
        },
        {
            'artifact_class_or_type_name':
                'SlicedClassificationMetrics',
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.SlicedClassificationMetrics',
                    schema_version='0.0.1')
        },
        {
            'artifact_class_or_type_name':
                artifact_types.SlicedClassificationMetrics,
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.SlicedClassificationMetrics',
                    schema_version='0.0.1')
        },
        {
            'artifact_class_or_type_name':
                'arbitrary name',
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.Artifact', schema_version='0.0.1')
        },
        {
            'artifact_class_or_type_name':
                _ArbitraryClass,
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.Artifact', schema_version='0.0.1')
        },
        {
            'artifact_class_or_type_name':
                artifact_types.HTML,
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.HTML', schema_version='0.0.1')
        },
        {
            'artifact_class_or_type_name':
                artifact_types.Markdown,
            'expected_result':
                pb.ArtifactTypeSchema(
                    schema_title='system.Markdown', schema_version='0.0.1')
        },
    )
    def test_get_artifact_type_schema(self, artifact_class_or_type_name,
                                      expected_result):
        self.assertEqual(
            expected_result,
            type_utils.get_artifact_type_schema(artifact_class_or_type_name))

    @parameterized.parameters(
        {
            'given_type': 'Int',
            'expected_type': pb.PrimitiveType.INT,
        },
        {
            'given_type': 'Integer',
            'expected_type': pb.PrimitiveType.INT,
        },
        {
            'given_type': int,
            'expected_type': pb.PrimitiveType.INT,
        },
        {
            'given_type': 'Double',
            'expected_type': pb.PrimitiveType.DOUBLE,
        },
        {
            'given_type': 'Float',
            'expected_type': pb.PrimitiveType.DOUBLE,
        },
        {
            'given_type': float,
            'expected_type': pb.PrimitiveType.DOUBLE,
        },
        {
            'given_type': 'String',
            'expected_type': pb.PrimitiveType.STRING,
        },
        {
            'given_type': 'Text',
            'expected_type': pb.PrimitiveType.STRING,
        },
        {
            'given_type': str,
            'expected_type': pb.PrimitiveType.STRING,
        },
        {
            'given_type': 'Boolean',
            'expected_type': pb.PrimitiveType.STRING,
        },
        {
            'given_type': bool,
            'expected_type': pb.PrimitiveType.STRING,
        },
        {
            'given_type': 'Dict',
            'expected_type': pb.PrimitiveType.STRING,
        },
        {
            'given_type': dict,
            'expected_type': pb.PrimitiveType.STRING,
        },
        {
            'given_type': 'List',
            'expected_type': pb.PrimitiveType.STRING,
        },
        {
            'given_type': list,
            'expected_type': pb.PrimitiveType.STRING,
        },
        {
            'given_type': Dict[str, int],
            'expected_type': pb.PrimitiveType.STRING,
        },
        {
            'given_type': List[Any],
            'expected_type': pb.PrimitiveType.STRING,
        },
        {
            'given_type': {
                'JsonObject': {
                    'data_type': 'proto:tfx.components.trainer.TrainArgs'
                }
            },
            'expected_type': pb.PrimitiveType.STRING,
        },
    )
    def test_get_parameter_type(self, given_type, expected_type):
        self.assertEqual(expected_type,
                         type_utils.get_parameter_type(given_type))

        # Test get parameter by Python type.
        self.assertEqual(pb.PrimitiveType.INT,
                         type_utils.get_parameter_type(int))

    def test_get_parameter_type_invalid(self):
        with self.assertRaises(AttributeError):
            type_utils.get_parameter_type_schema(None)

    def test_get_input_artifact_type_schema(self):
        input_specs = [
            structures.InputSpec(name='input1', type='String'),
            structures.InputSpec(name='input2', type='Model'),
            structures.InputSpec(name='input3', type=None),
        ]
        # input not found.
        with self.assertRaises(AssertionError) as cm:
            type_utils.get_input_artifact_type_schema('input0', input_specs)
            self.assertEqual('Input not found.', str(cm))

        # input found, but it doesn't map to an artifact type.
        with self.assertRaises(AssertionError) as cm:
            type_utils.get_input_artifact_type_schema('input1', input_specs)
            self.assertEqual('Input is not an artifact type.', str(cm))

        # input found, and a matching artifact type schema returned.
        self.assertEqual(
            'system.Model',
            type_utils.get_input_artifact_type_schema('input2',
                                                      input_specs).schema_title)

        # input found, and the default artifact type schema returned.
        self.assertEqual(
            'system.Artifact',
            type_utils.get_input_artifact_type_schema('input3',
                                                      input_specs).schema_title)

    def test_get_parameter_type_field_name(self):
        self.assertEqual('string_value',
                         type_utils.get_parameter_type_field_name('String'))
        self.assertEqual('int_value',
                         type_utils.get_parameter_type_field_name('Integer'))
        self.assertEqual('double_value',
                         type_utils.get_parameter_type_field_name('Float'))


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.v2.components.experimental.base_component."""

import unittest
from unittest.mock import patch

from kfp.v2.components.experimental import base_component
from kfp.v2.components.experimental import component_spec
from kfp.v2.components.experimental import pipeline_task


class TestComponent(base_component.BaseComponent):

    def execute(self, *args, **kwargs):
        pass


component_op = TestComponent(
    component_spec=component_spec.ComponentSpec(
        name='component_1',
        implementation=component_spec.ContainerSpec(
            image='alpine',
            commands=[
                'sh',
                '-c',
                'set -ex\necho "$0" "$1" "$2" > "$3"',
                component_spec.InputValuePlaceholder(name='input1'),
                component_spec.InputValuePlaceholder(name='input2'),
                component_spec.InputValuePlaceholder(name='input3'),
                component_spec.OutputPathPlaceholder(name='output1'),
            ],
        ),
        inputs={
            'input1': component_spec.InputSpec(type='String'),
            'input2': component_spec.InputSpec(type='Integer'),
            'input3': component_spec.InputSpec(type='Float', default=3.14),
        },
        outputs={
            'output1': component_spec.OutputSpec(name='output1', type='String'),
        },
    ))


class BaseComponentTest(unittest.TestCase):

    @patch.object(pipeline_task, 'create_pipeline_task', autospec=True)
    def test_instantiate_component_with_keyword_arguments(
            self, mock_create_pipeline_task):

        component_op(input1='hello', input2=100, input3=1.23)

        mock_create_pipeline_task.assert_called_once_with(
            component_spec=component_op.component_spec,
            arguments={
                'input1': 'hello',
                'input2': 100,
                'input3': 1.23,
            })

    @patch.object(pipeline_task, 'create_pipeline_task', autospec=True)
    def test_instantiate_component_omitting_arguments_with_default(
            self, mock_create_pipeline_task):

        component_op(input1='hello', input2=100)

        mock_create_pipeline_task.assert_called_once_with(
            component_spec=component_op.component_spec,
            arguments={
                'input1': 'hello',
                'input2': 100,
                'input3': '3.14',
            })

    def test_instantiate_component_with_positional_arugment(self):
        with self.assertRaisesRegex(
                TypeError,
                'Components must be instantiated using keyword arguments. Positional '
                'parameters are not allowed \(found 3 such parameters for component '
                '"component_1"\).'):
            component_op('abc', 1, 2.3)

    def test_instantiate_component_with_unexpected_keyword_arugment(self):
        with self.assertRaisesRegex(
                TypeError,
                'component_1\(\) got an unexpected keyword argument "input4".'):
            component_op(input1='abc', input2=1, input3=2.3, input4='extra')

    def test_instantiate_component_with_missing_arugments(self):
        with self.assertRaisesRegex(
                TypeError,
                'component_1\(\) missing 1 required positional argument: input1.'
        ):
            component_op(input2=1)

        with self.assertRaisesRegex(
                TypeError,
                'component_1\(\) missing 2 required positional arguments: input1,input2.'
        ):
            component_op()


if __name__ == '__main__':
    unittest.main()



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""KFP v2 DSL compiler utility functions."""

import re
from typing import Any, Mapping, Optional, Union

from kfp.containers import _component_builder
from kfp.dsl import _container_op
from kfp.dsl import _pipeline_param
from kfp.pipeline_spec import pipeline_spec_pb2
from kfp.v2.components.types import type_utils

# Alias for PipelineContainerSpec
PipelineContainerSpec = pipeline_spec_pb2.PipelineDeploymentConfig.PipelineContainerSpec


def build_runtime_config_spec(
    output_directory: str,
    pipeline_parameters: Optional[Mapping[
        str, _pipeline_param.PipelineParam]] = None,
) -> pipeline_spec_pb2.PipelineJob.RuntimeConfig:
    """Converts pipeine parameters to runtime parameters mapping.

    Args:
      output_directory: The root of pipeline outputs.
      pipeline_parameters: The mapping from parameter names to PipelineParam
        objects. Optional.

    Returns:
      A pipeline job RuntimeConfig object.
    """

    def _get_value(
            param: _pipeline_param.PipelineParam) -> pipeline_spec_pb2.Value:
        assert param.value is not None, 'None values should be filterd out.'

        result = pipeline_spec_pb2.Value()
        # TODO(chensun): remove defaulting to 'String' for None param_type once we
        # fix importer behavior.
        param_type = type_utils.get_parameter_type(param.param_type or 'String')
        if param_type == pipeline_spec_pb2.PrimitiveType.INT:
            result.int_value = int(param.value)
        elif param_type == pipeline_spec_pb2.PrimitiveType.DOUBLE:
            result.double_value = float(param.value)
        elif param_type == pipeline_spec_pb2.PrimitiveType.STRING:
            result.string_value = str(param.value)
        else:
            # For every other type, defaults to 'String'.
            # TODO(chensun): remove this default behavior once we migrate from
            # `pipeline_spec_pb2.Value` to `protobuf.Value`.
            result.string_value = str(param.value)

        return result

    parameters = pipeline_parameters or {}
    return pipeline_spec_pb2.PipelineJob.RuntimeConfig(
        gcs_output_directory=output_directory,
        parameters={
            k: _get_value(v)
            for k, v in parameters.items()
            if v.value is not None
        })


def validate_pipeline_name(name: str) -> None:
    """Validate pipeline name.

    A valid pipeline name should match ^[a-z0-9][a-z0-9-]{0,127}$.

    Args:
      name: The pipeline name.

    Raises:
      ValueError if the pipeline name doesn't conform to the regular expression.
    """
    pattern = re.compile(r'^[a-z0-9][a-z0-9-]{0,127}$')
    if not pattern.match(name):
        raise ValueError(
            'Invalid pipeline name: %s.\n'
            'Please specify a pipeline name that matches the regular '
            'expression "^[a-z0-9][a-z0-9-]{0,127}$" using '
            '`dsl.pipeline(name=...)` decorator.' % name)


def is_v2_component(op: _container_op.ContainerOp) -> bool:
    """Determines whether a component is a KFP v2 component."""
    component_spec = op._metadata
    return (component_spec and component_spec.metadata and
            component_spec.metadata.annotations and
            component_spec.metadata.annotations.get(
                _component_builder.V2_COMPONENT_ANNOTATION) == 'true')


def refactor_v2_container_spec(container_spec: PipelineContainerSpec) -> None:
    """Refactor the container spec for a v2 component."""
    if not '--function_name' in container_spec.args:
        raise RuntimeError(
            'V2 component is expected to have function_name as a '
            'command line arg.')
    fn_name_idx = list(container_spec.args).index('--function_name') + 1
    fn_name = container_spec.args[fn_name_idx]
    container_spec.ClearField('command')
    container_spec.ClearField('args')
    container_spec.command.extend(['python', '-m', 'kfp.container.entrypoint'])
    container_spec.args.extend(
        ['--executor_input_str', '{{$}}', '--function_name', fn_name])



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definitions of utils methods."""

import re


def maybe_rename_for_k8s(name: str) -> str:
    """Cleans and converts a name to be k8s compatible.

    Args:
      name: The original name.

    Returns:
      A sanitized name.
    """
    return re.sub('-+', '-', re.sub('[^-0-9a-z]+', '-',
                                    name.lower())).lstrip('-').rstrip('-')



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.v2.google.client._cloud_function_create_pipeline_job."""

import datetime
import json
import os
import unittest

from kfp.v2.google.client import _cloud_function_templated_http_request


class CloudFunctionCreatePipelineJobTest(unittest.TestCase):

    def test_preprocess(self):
        test_data_path = os.path.join(
            os.path.dirname(__file__),
            'testdata',
        )
        function_request_path = os.path.join(
            test_data_path,
            'pipeline1_request_body.json',
        )
        expected_pipeline_request_path = os.path.join(
            test_data_path,
            'pipeline1_request_body_final.json',
        )

        with open(function_request_path, 'rb') as f:
            function_request = f.read()

        with open(expected_pipeline_request_path, 'r') as f:
            expected_pipeline_request = json.load(f)

        (_, _, resolved_request_body
        ) = _cloud_function_templated_http_request._preprocess_request_body(
            function_request, time=datetime.datetime(2020, 8, 1, 12, 34))
        actual_pipeline_request = json.loads(resolved_request_body)
        self.assertEqual(actual_pipeline_request, expected_pipeline_request)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Builder for CAIP pipelines Pipeline level proto spec."""

import copy
import json
from typing import Any, Dict, Mapping, Optional, Union


class RuntimeConfigBuilder(object):
    """CAIP pipelines RuntimeConfig builder.

    Constructs a RuntimeConfig spec with pipeline_root and parameter
    overrides.
    """

    def __init__(
        self,
        pipeline_root: str,
        parameter_types: Mapping[str, str],
        parameter_values: Optional[Dict[str, Any]] = None,
    ):
        """Creates a RuntimeConfigBuilder object.

        Args:
          pipeline_root: The root of the pipeline outputs.
          parameter_types: The mapping from pipeline parameter name to its type.
          parameter_values: The mapping from runtime parameter name to its value.
        """
        self._pipeline_root = pipeline_root
        self._parameter_types = parameter_types
        self._parameter_values = copy.deepcopy(parameter_values or {})

    @classmethod
    def from_job_spec_json(
            cls, job_spec: Mapping[str, Any]) -> 'RuntimeConfigBuilder':
        """Creates a RuntimeConfigBuilder object from PipelineJob json spec.

        Args:
          job_spec: The PipelineJob spec.

        Returns:
          A RuntimeConfigBuilder object.
        """
        runtime_config_spec = job_spec['runtimeConfig']
        parameter_types = {}
        parameter_input_definitions = job_spec['pipelineSpec']['root'].get(
            'inputDefinitions', {}).get('parameters', {})
        for k, v in parameter_input_definitions.items():
            parameter_types[k] = v['type']

        pipeline_root = runtime_config_spec.get('gcsOutputDirectory')
        parameter_values = _parse_runtime_parameters(runtime_config_spec)
        return cls(pipeline_root, parameter_types, parameter_values)

    def update_pipeline_root(self, pipeline_root: Optional[str]) -> None:
        """Updates pipeline_root value.

        Args:
          pipeline_root: The root of the pipeline outputs.
        """
        if pipeline_root:
            self._pipeline_root = pipeline_root

    def update_runtime_parameters(
            self, parameter_values: Optional[Mapping[str, Any]]) -> None:
        """Merges runtime parameter values.

        Args:
          parameter_values: The mapping from runtime parameter names to its values.
        """
        if parameter_values:
            for k, v in parameter_values.items():
                if isinstance(v, (dict, list, bool)):
                    parameter_values[k] = json.dumps(v)
        if parameter_values:
            self._parameter_values.update(parameter_values)

    def build(self) -> Mapping[str, Any]:
        """Build a RuntimeConfig proto."""
        if not self._pipeline_root:
            raise ValueError(
                'Pipeline root must be specified, either during compile '
                'time, or when calling the service.')
        return {
            'gcsOutputDirectory': self._pipeline_root,
            'parameters': {
                k: self._get_caip_value(k, v)
                for k, v in self._parameter_values.items()
                if v is not None
            }
        }

    def _get_caip_value(self, name: str,
                        value: Union[int, float, str]) -> Mapping[str, Any]:
        """Converts primitive values into CAIP pipeline Value proto message.

        Args:
          name: The name of the pipeline parameter.
          value: The value of the pipeline parameter.

        Returns:
          A dictionary represents the CAIP pipeline Value proto message.

        Raises:
          AssertionError: if the value is None.
          ValueError: if the parameeter name is not found in pipeline root inputs.
          TypeError: if the paraemter type is not one of 'INT', 'DOUBLE', 'STRING'.
        """
        assert value is not None, 'None values should be filterd out.'

        if name not in self._parameter_types:
            raise ValueError(
                'The pipeline parameter {} is not found in the pipeline '
                'job input definitions.'.format(name))

        result = {}
        if self._parameter_types[name] == 'INT':
            result['intValue'] = value
        elif self._parameter_types[name] == 'DOUBLE':
            result['doubleValue'] = value
        elif self._parameter_types[name] == 'STRING':
            result['stringValue'] = value
        else:
            raise TypeError('Got unknown type of value: {}'.format(value))

        return result


def _parse_runtime_parameters(
        runtime_config_spec: Mapping[str, Any]) -> Optional[Dict[str, Any]]:
    """Extracts runtime parameters from runtime config json spec."""
    runtime_parameters = runtime_config_spec.get('parameters')
    if not runtime_parameters:
        return None

    result = {}
    for name, value in runtime_parameters.items():
        if 'intValue' in value:
            result[name] = int(value['intValue'])
        elif 'doubleValue' in value:
            result[name] = float(value['doubleValue'])
        elif 'stringValue' in value:
            result[name] = value['stringValue']
        else:
            raise TypeError('Got unknown type of value: {}'.format(value))

    return result



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import json
from typing import Any, Callable, Dict, List, Optional, Union

from kfp.v2.components.types import artifact_types, type_annotations


class Executor():
    """Executor executes v2-based Python function components."""

    def __init__(self, executor_input: Dict, function_to_execute: Callable):
        self._func = function_to_execute
        self._input = executor_input
        self._input_artifacts: Dict[str, artifact_types.Artifact] = {}
        self._output_artifacts: Dict[str, artifact_types.Artifact] = {}

        for name, artifacts in self._input.get('inputs',
                                               {}).get('artifacts', {}).items():
            artifacts_list = artifacts.get('artifacts')
            if artifacts_list:
                self._input_artifacts[name] = self._make_input_artifact(
                    artifacts_list[0])

        for name, artifacts in self._input.get('outputs',
                                               {}).get('artifacts', {}).items():
            artifacts_list = artifacts.get('artifacts')
            if artifacts_list:
                self._output_artifacts[name] = self._make_output_artifact(
                    artifacts_list[0])

        self._return_annotation = inspect.signature(
            self._func).return_annotation
        self._executor_output = {}

    @classmethod
    def _make_input_artifact(cls, runtime_artifact: Dict):
        return artifact_types.create_runtime_artifact(runtime_artifact)

    @classmethod
    def _make_output_artifact(cls, runtime_artifact: Dict):
        import os
        artifact = artifact_types.create_runtime_artifact(runtime_artifact)
        os.makedirs(os.path.dirname(artifact.path), exist_ok=True)
        return artifact

    def _get_input_artifact(self, name: str):
        return self._input_artifacts.get(name)

    def _get_output_artifact(self, name: str):
        return self._output_artifacts.get(name)

    def _get_input_parameter_value(self, parameter_name: str,
                                   parameter_type: Any):
        parameter = self._input.get('inputs',
                                    {}).get('parameters',
                                            {}).get(parameter_name, None)
        if parameter is None:
            return None

        if parameter.get('stringValue') is not None:
            if parameter_type == str:
                return parameter['stringValue']
            elif parameter_type == bool:
                # Use `.lower()` so it can also handle 'True' and 'False' (resulted from
                # `str(True)` and `str(False)`, respectively.
                return json.loads(parameter['stringValue'].lower())
            else:
                return json.loads(parameter['stringValue'])
        elif parameter.get('intValue'):
            return int(parameter['intValue'])
        elif parameter.get('doubleValue'):
            return float(parameter['doubleValue'])

    def _get_output_parameter_path(self, parameter_name: str):
        parameter = self._input.get('outputs',
                                    {}).get('parameters',
                                            {}).get(parameter_name, None)
        if parameter is None:
            return None

        import os
        path = parameter.get('outputFile', None)
        if path:
            os.makedirs(os.path.dirname(path), exist_ok=True)
        return path

    def _get_output_artifact_path(self, artifact_name: str):
        output_artifact = self._output_artifacts.get(artifact_name)
        if not output_artifact:
            raise ValueError(
                'Failed to get output artifact path for artifact name {}'
                .format(artifact_name))
        return output_artifact.path

    def _get_input_artifact_path(self, artifact_name: str):
        input_artifact = self._input_artifacts.get(artifact_name)
        if not input_artifact:
            raise ValueError(
                'Failed to get input artifact path for artifact name {}'.format(
                    artifact_name))
        return input_artifact.path

    def _write_output_parameter_value(self, name: str,
                                      value: Union[str, int, float, bool, dict,
                                                   list, Dict, List]):
        if type(value) == str:
            output = {'stringValue': value}
        elif type(value) == int:
            output = {'intValue': value}
        elif type(value) == float:
            output = {'doubleValue': value}
        else:
            # For bool, list, dict, List, Dict, json serialize the value.
            output = {'stringValue': json.dumps(value)}

        if not self._executor_output.get('parameters'):
            self._executor_output['parameters'] = {}

        self._executor_output['parameters'][name] = output

    def _write_output_artifact_payload(self, name: str, value: Any):
        path = self._get_output_artifact_path(name)
        with open(path, 'w') as f:
            f.write(str(value))

    # TODO: extract to a util
    @classmethod
    def _get_short_type_name(cls, type_name: str) -> str:
        """Extracts the short form type name.

        This method is used for looking up serializer for a given type.

        For example:
          typing.List -> List
          typing.List[int] -> List
          typing.Dict[str, str] -> Dict
          List -> List
          str -> str

        Args:
          type_name: The original type name.

        Returns:
          The short form type name or the original name if pattern doesn't match.
        """
        import re
        match = re.match('(typing\.)?(?P<type>\w+)(?:\[.+\])?', type_name)
        if match:
            return match.group('type')
        else:
            return type_name

    # TODO: merge with type_utils.is_parameter_type
    @classmethod
    def _is_parameter(cls, annotation: Any) -> bool:
        if type(annotation) == type:
            return annotation in [str, int, float, bool, dict, list]

        # Annotation could be, for instance `typing.Dict[str, str]`, etc.
        return cls._get_short_type_name(str(annotation)) in ['Dict', 'List']

    @classmethod
    def _is_artifact(cls, annotation: Any) -> bool:
        if type(annotation) == type:
            return issubclass(annotation, artifact_types.Artifact)
        return False

    @classmethod
    def _is_named_tuple(cls, annotation: Any) -> bool:
        if type(annotation) == type:
            return issubclass(annotation, tuple) and hasattr(
                annotation, '_fields') and hasattr(annotation,
                                                   '__annotations__')
        return False

    def _handle_single_return_value(self, output_name: str,
                                    annotation_type: Any, return_value: Any):
        if self._is_parameter(annotation_type):
            origin_type = getattr(annotation_type, '__origin__', None) or annotation_type
            if not isinstance(return_value, origin_type):
                raise ValueError(
                    'Function `{}` returned value of type {}; want type {}'
                    .format(self._func.__name__, type(return_value),
                            origin_type))
            self._write_output_parameter_value(output_name, return_value)
        elif self._is_artifact(annotation_type):
            self._write_output_artifact_payload(output_name, return_value)
        else:
            raise RuntimeError(
                'Unknown return type: {}. Must be one of `str`, `int`, `float`, or a'
                ' subclass of `Artifact`'.format(annotation_type))

    def _write_executor_output(self, func_output: Optional[Any] = None):
        if self._output_artifacts:
            self._executor_output['artifacts'] = {}

        for name, artifact in self._output_artifacts.items():
            runtime_artifact = {
                'name': artifact.name,
                'uri': artifact.uri,
                'metadata': artifact.metadata,
            }
            artifacts_list = {'artifacts': [runtime_artifact]}

            self._executor_output['artifacts'][name] = artifacts_list

        if func_output is not None:
            if self._is_parameter(self._return_annotation) or self._is_artifact(
                    self._return_annotation):
                # Note: single output is named `Output` in component.yaml.
                self._handle_single_return_value('Output',
                                                 self._return_annotation,
                                                 func_output)
            elif self._is_named_tuple(self._return_annotation):
                if len(self._return_annotation._fields) != len(func_output):
                    raise RuntimeError(
                        'Expected {} return values from function `{}`, got {}'
                        .format(
                            len(self._return_annotation._fields),
                            self._func.__name__, len(func_output)))
                for i in range(len(self._return_annotation._fields)):
                    field = self._return_annotation._fields[i]
                    field_type = self._return_annotation.__annotations__[field]
                    if type(func_output) == tuple:
                        field_value = func_output[i]
                    else:
                        field_value = getattr(func_output, field)
                    self._handle_single_return_value(field, field_type,
                                                     field_value)
            else:
                raise RuntimeError(
                    'Unknown return type: {}. Must be one of `str`, `int`, `float`, a'
                    ' subclass of `Artifact`, or a NamedTuple collection of these types.'
                    .format(self._return_annotation))

        import os
        os.makedirs(
            os.path.dirname(self._input['outputs']['outputFile']),
            exist_ok=True)
        with open(self._input['outputs']['outputFile'], 'w') as f:
            f.write(json.dumps(self._executor_output))

    def execute(self):
        annotations = inspect.getfullargspec(self._func).annotations

        # Function arguments.
        func_kwargs = {}

        for k, v in annotations.items():
            if k == 'return':
                continue

            # Annotations for parameter types could be written as, for example,
            # `Optional[str]`. In this case, we need to strip off the part
            # `Optional[]` to get the actual parameter type.
            v = type_annotations.maybe_strip_optional_from_annotation(v)

            if self._is_parameter(v):
                func_kwargs[k] = self._get_input_parameter_value(k, v)

            if type_annotations.is_artifact_annotation(v):
                if type_annotations.is_input_artifact(v):
                    func_kwargs[k] = self._get_input_artifact(k)
                if type_annotations.is_output_artifact(v):
                    func_kwargs[k] = self._get_output_artifact(k)

            elif isinstance(v, type_annotations.OutputPath):
                if self._is_parameter(v.type):
                    func_kwargs[k] = self._get_output_parameter_path(k)
                else:
                    func_kwargs[k] = self._get_output_artifact_path(k)
            elif isinstance(v, type_annotations.InputPath):
                func_kwargs[k] = self._get_input_artifact_path(k)

        result = self._func(**func_kwargs)
        self._write_executor_output(result)



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for scheduling pipeline runs."""

import base64
import hashlib
import json
import logging
import pathlib
import re
import tempfile
from typing import Any, Mapping, Optional
import zipfile

import googleapiclient
from googleapiclient import discovery
import requests

from kfp.v2.google.client import client_utils
from kfp.v2.google.client import runtime_config_builder

_PROXY_FUNCTION_NAME = 'templated_http_request-v1'
_PROXY_FUNCTION_FILENAME = '_cloud_function_templated_http_request.py'

_CAIPP_ENDPOINT_WITHOUT_REGION = 'aiplatform.googleapis.com'
_CAIPP_API_VERSION = 'v1beta1'


def create_from_pipeline_file(
    pipeline_path: str,
    schedule: str,
    project_id: str,
    region: str = 'us-central1',
    time_zone: str = 'US/Pacific',
    parameter_values: Optional[Mapping[str, Any]] = None,
    pipeline_root: Optional[str] = None,
    service_account: Optional[str] = None,
    app_engine_region: Optional[str] = None,
    cloud_scheduler_service_account: Optional[str] = None,
) -> dict:
    """Creates schedule for compiled pipeline file.

    This function creates scheduled job which will run the provided pipeline on
    schedule. This is implemented by creating a Google Cloud Scheduler Job.
    The job will be visible in https://console.google.com/cloudscheduler and can
    be paused/resumed and deleted.

    To make the system work, this function also creates a Google Cloud Function
    which acts as an intermediary between the Scheduler and Pipelines. A single
    function is shared between all scheduled jobs.
    The following APIs will be activated automatically:
    * cloudfunctions.googleapis.com
    * cloudscheduler.googleapis.com
    * appengine.googleapis.com

    Args:
      pipeline_path: Path of the compiled pipeline file.
      schedule: Schedule in cron format. Example: "45 * * * *"
      project_id: Google Cloud project ID
      region: Google Cloud compute region. Default is 'us-central1'
      time_zone: Schedule time zone. Default is 'US/Pacific'
      parameter_values: Arguments for the pipeline parameters
      pipeline_root: Optionally the user can override the pipeline root
        specified during the compile time.
      service_account: The service account that the pipeline workload runs as.
      app_engine_region: The region that cloud scheduler job is created in.
      cloud_scheduler_service_account: The service account that Cloud Scheduler job and the proxy cloud function use.
        this should have permission to call AI Platform API and the proxy function.
        If not specified, the functions uses the App Engine default service account.

    Returns:
      Created Google Cloud Scheduler Job object dictionary.
    """
    pipeline_dict = client_utils.load_json(pipeline_path)

    return _create_from_pipeline_dict(
        pipeline_dict=pipeline_dict,
        schedule=schedule,
        project_id=project_id,
        region=region,
        time_zone=time_zone,
        parameter_values=parameter_values,
        pipeline_root=pipeline_root,
        service_account=service_account,
        app_engine_region=app_engine_region,
        cloud_scheduler_service_account=cloud_scheduler_service_account,
    )


def _create_from_pipeline_dict(
    pipeline_dict: dict,
    schedule: str,
    project_id: str,
    region: str = 'us-central1',
    time_zone: str = 'US/Pacific',
    parameter_values: Optional[Mapping[str, Any]] = None,
    pipeline_root: Optional[str] = None,
    service_account: Optional[str] = None,
    app_engine_region: Optional[str] = None,
    cloud_scheduler_service_account: Optional[str] = None,
) -> dict:
    """Creates schedule for compiled pipeline dictionary."""

    _enable_required_apis(project_id=project_id)

    # If appengine region is not provided, use the pipeline region.
    app_engine_region = app_engine_region or region

    proxy_function_url = _get_proxy_cloud_function_endpoint(
        project_id=project_id,
        region=region,
        cloud_scheduler_service_account=cloud_scheduler_service_account,
    )

    if parameter_values or pipeline_root:
        config_builder = runtime_config_builder.RuntimeConfigBuilder.from_job_spec_json(
            pipeline_dict)
        config_builder.update_runtime_parameters(
            parameter_values=parameter_values)
        config_builder.update_pipeline_root(pipeline_root=pipeline_root)
        updated_runtime_config = config_builder.build()
        pipeline_dict['runtimeConfig'] = updated_runtime_config

    # Creating job creation request to get the final request URL
    pipeline_jobs_api_url = f'https://{region}-{_CAIPP_ENDPOINT_WITHOUT_REGION}/{_CAIPP_API_VERSION}/projects/{project_id}/locations/{region}/pipelineJobs'

    # Preparing the request body for the Cloud Function processing
    full_pipeline_name = pipeline_dict.get('name')
    pipeline_display_name = pipeline_dict.get('displayName')
    time_format_suffix = "-{{$.scheduledTime.strftime('%Y-%m-%d-%H-%M-%S')}}"
    if 'name' in pipeline_dict:
        pipeline_dict['name'] += time_format_suffix
    if 'displayName' in pipeline_dict:
        pipeline_dict['displayName'] += time_format_suffix

    pipeline_dict['_url'] = pipeline_jobs_api_url
    pipeline_dict['_method'] = 'POST'

    if service_account is not None:
        pipeline_dict['serviceAccount'] = service_account

    pipeline_text = json.dumps(pipeline_dict)
    pipeline_data = pipeline_text.encode('utf-8')

    # Generating human-readable schedule name.
    schedule_name = _build_schedule_name(
        job_body_data=pipeline_data,
        schedule=schedule,
        pipeline_name=full_pipeline_name,
        display_name=pipeline_display_name,
    )

    project_location_path = 'projects/{}/locations/{}'.format(
        project_id, app_engine_region)
    scheduled_job_full_name = '{}/jobs/{}'.format(project_location_path,
                                                  schedule_name)
    service_account_email = cloud_scheduler_service_account or '{}@appspot.gserviceaccount.com'.format(
        project_id)

    scheduled_job = dict(
        name=scheduled_job_full_name,  # Optional. Only used for readable names.
        schedule=schedule,
        time_zone=time_zone,
        http_target=dict(
            http_method='POST',
            uri=proxy_function_url,
            # Warning: when using google.cloud.scheduler_v1, the type of body is
            # bytes or string. But when using the API through discovery, the body
            # needs to be base64-encoded.
            body=base64.b64encode(pipeline_data).decode('utf-8'),
            oidc_token=dict(service_account_email=service_account_email,),
        ),
        # TODO(avolkov): Add labels once Cloud Scheduler supports them
        # labels={
        #     'google.cloud.ai-platform.pipelines.scheduling': 'v1alpha1',
        # },
    )

    try:
        response = _create_scheduler_job(
            project_location_path=project_location_path,
            job_body=scheduled_job,
        )
        return response
    except googleapiclient.errors.HttpError as err:
        # Handling the case where the exact schedule already exists.
        if err.resp.get('status') == '409':
            raise RuntimeError(
                'The exact same schedule already exists') from err
        raise err


def _create_scheduler_job(project_location_path: str,
                          job_body: Mapping[str, Any]) -> str:
    """Creates a scheduler job.

    Args:
      project_location_path: The project location path.
      job_body: The scheduled job dictionary object.

    Returns:
      The response from scheduler service.
    """
    # We cannot use google.cloud.scheduler_v1.CloudSchedulerClient since
    # it's not available internally.
    scheduler_service = discovery.build(
        'cloudscheduler', 'v1', cache_discovery=False)
    scheduler_jobs_api = scheduler_service.projects().locations().jobs()
    response = scheduler_jobs_api.create(
        parent=project_location_path,
        body=job_body,
    ).execute()
    return response


def _build_schedule_name(
    job_body_data: bytes,
    schedule: str,
    pipeline_name: str,
    display_name: str,
) -> str:
    """Generates the name for the schedule.

    Args:
      job_body_data: The serialized pipeline job.
      schedule: Schedule in cron format.
      pipeline_name: Full resource name of the pipeline in
        projects/<project>/pipelineJobs/<pipeline_id> format.
      display_name: Pipeline display name.
    Returns:
      Suggested schedule resource name.
    """
    pipeline_name_part = 'pipeline'
    if pipeline_name is not None:
        # pipeline_name format: projects/<project>/pipelineJobs/<pipeline_id>
        pipeline_id = pipeline_name.split('/')[-1]
        # Limiting the length of the pipeline name part.
        pipeline_name_part = pipeline_id[0:200]
    elif display_name is not None:
        pipeline_name_part = display_name
    pipeline_hash_part = hashlib.sha256(job_body_data).hexdigest()[0:8]
    schedule_part = (
        schedule.replace('*/', 'div').replace('*', 'a').replace(' ', '-'))
    job_name = '_'.join([
        'pipeline',
        pipeline_name_part,
        pipeline_hash_part,
        schedule_part,
    ])
    job_name = re.sub('[^-_a-z0-9]', '_', job_name)
    return job_name


# For mocking
def _get_cloud_functions_api():
    functions_service = discovery.build(
        'cloudfunctions', 'v1', cache_discovery=False)
    functions_api = functions_service.projects().locations().functions()
    return functions_api


def _create_or_get_cloud_function(
    name: str,
    entry_point: str,
    file_data: Mapping[str, str],
    project_id: str,
    region: str,
    runtime: str = 'python37',
    cloud_scheduler_service_account: Optional[str] = None,
):
    """Creates Google Cloud Function."""
    functions_api = _get_cloud_functions_api()

    project_location_path = 'projects/{}/locations/{}'.format(
        project_id, region)
    function_full_name = project_location_path + '/functions/' + name

    # Returning early if the function already exists.
    try:
        function_get_response = functions_api.get(
            name=function_full_name).execute()
        return function_get_response
    except googleapiclient.errors.HttpError as err:
        raise_error = True
        if err.resp['status'] == '404':
            # The function does not exist, which is expected.
            raise_error = False
        if raise_error:
            raise err

    get_upload_url_response = functions_api.generateUploadUrl(
        parent=project_location_path,
        body={},
    ).execute()
    upload_url = get_upload_url_response['uploadUrl']

    # Preparing the payload archive
    with tempfile.TemporaryFile() as archive_file:
        with zipfile.ZipFile(archive_file, 'w',
                             zipfile.ZIP_DEFLATED) as archive:
            for path, data in file_data.items():
                archive.writestr(path, data)

        archive_file.seek(0)
        headers = {
            'content-type': 'application/zip',
            'x-goog-content-length-range': '0,104857600',
        }
        upload_response = requests.put(
            url=upload_url,
            headers=headers,
            data=archive_file,
        )
        upload_response.raise_for_status()

    # Prepare Request Body
    # https://cloud.google.com/functions/docs/reference/rest/v1/projects.locations.functions#resource-cloudfunction

    request_body = {
        'name': function_full_name,
        'entryPoint': entry_point,
        'sourceUploadUrl': upload_url,
        'httpsTrigger': {},
        'runtime': runtime,
    }
    if cloud_scheduler_service_account is not None:
        request_body["serviceAccountEmail"] = cloud_scheduler_service_account
    try:
        functions_api.create(
            location=project_location_path,
            body=request_body,
        ).execute()
        # Response is an operation object dict
        # TODO(avolkov): Wait for the operation to succeed
        # 'status' can be 'ACTIVE', 'DEPLOY_IN_PROGRESS', etc
    except googleapiclient.errors.HttpError as err:
        # Handling the case where the function already exists
        raise_error = True
        if err.resp['status'] == '409':
            err_content_dict = json.loads(err.content)
            err_error_dict = err_content_dict.get('error')
            if err_error_dict and err_error_dict.get(
                    'status') == 'ALREADY_EXISTS':
                # This should not usually happen
                logging.warning(
                    'Cloud Function already exists: name=%s',
                    function_full_name,
                )
                raise_error = False
        if raise_error:
            raise err

    function_get_response = functions_api.get(name=function_full_name).execute()
    logging.info('Created Cloud Function: name=%s', function_full_name)

    return function_get_response


def _enable_required_apis(project_id: str,):
    """Enables necessary APIs."""
    serviceusage_service = discovery.build(
        'serviceusage', 'v1', cache_discovery=False)
    services_api = serviceusage_service.services()

    required_services = [
        'cloudfunctions.googleapis.com',
        'cloudscheduler.googleapis.com',
        'appengine.googleapis.com',  # Required by the Cloud Scheduler.
    ]
    project_path = 'projects/' + project_id
    for service_name in required_services:
        service_path = project_path + '/services/' + service_name
        services_api.enable(name=service_path).execute()


def _get_proxy_cloud_function_endpoint(
    project_id: str,
    region: str = 'us-central1',
    cloud_scheduler_service_account: Optional[str] = None,
):
    """Sets up a proxy Cloud Function."""
    function_source_path = (
        pathlib.Path(__file__).parent / _PROXY_FUNCTION_FILENAME)
    function_source = function_source_path.read_text()
    function_entry_point = '_process_request'

    function_dict = _create_or_get_cloud_function(
        name=_PROXY_FUNCTION_NAME,
        entry_point=function_entry_point,
        project_id=project_id,
        region=region,
        runtime='python37',
        file_data={
            'main.py': function_source,
            'requirements.txt': 'google-api-python-client>=1.7.8,<2',
        },
        cloud_scheduler_service_account=cloud_scheduler_service_account,
    )
    endpoint_url = function_dict['httpsTrigger']['url']
    return endpoint_url



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.v2.google.client.schedule."""

import base64
import hashlib
import json
import os
import unittest
from unittest import mock

from kfp.v2.google.client import schedule


class ScheduleTest(unittest.TestCase):

    def test_create_from_pipeline_file(self):
        test_data_path = os.path.join(os.path.dirname(__file__), 'testdata')
        pipeline_path = os.path.join(test_data_path, 'pipeline1.json')
        pipeline_request_body_path = os.path.join(
            test_data_path, 'pipeline1_request_body.json')

        project_id = 'project-id'
        location = 'us-central1'
        function_url = ('https://{}-{}.cloudfunctions.net/' +
                        'templated_http_request-v1').format(
                            location, project_id)
        with mock.patch(
                'kfp.v2.google.client.schedule._enable_required_apis',
                return_value=None,
        ), mock.patch(
                'kfp.v2.google.client.schedule._get_proxy_cloud_function_endpoint',
                return_value=function_url,
        ), mock.patch(
                'kfp.v2.google.client.schedule._create_scheduler_job',
                spec=True) as create_scheduler_job_mock:
            schedule.create_from_pipeline_file(
                pipeline_path=pipeline_path,
                schedule='46 * * * *',
                project_id=project_id,
                region=location,
                time_zone='America/Los_Angeles',
                parameter_values={'name_param': 'World'},
                pipeline_root='gs://my-project/pipeline_root/tmp/',
            )

            with open(pipeline_request_body_path, 'rb') as f:
                expected_body_dict = json.load(f)
            expected_body_json = json.dumps(expected_body_dict)
            expected_body_data = expected_body_json.encode('utf-8')
            expected_body_data_hash = hashlib.sha256(
                expected_body_data).hexdigest()[0:8]

            create_scheduler_job_mock.assert_called_with(
                project_location_path='projects/{}/locations/{}'.format(
                    project_id, location),
                job_body={
                    'name':
                        'projects/{}/locations/{}/jobs/pipeline_my-pipeline_{}_46-a-a-a-a'
                        .format(project_id, location, expected_body_data_hash),
                    'schedule':
                        '46 * * * *',
                    'time_zone':
                        'America/Los_Angeles',
                    'http_target': {
                        'http_method':
                            'POST',
                        'uri':
                            function_url,
                        'body':
                            base64.b64encode(expected_body_data).decode('utf-8'
                                                                       ),
                        'oidc_token': {
                            'service_account_email':
                                'project-id@appspot.gserviceaccount.com',
                        },
                    },
                },
            )

    def test_create_schedule_when_cloud_function_already_exists(self):
        test_data_path = os.path.join(os.path.dirname(__file__), 'testdata')
        pipeline_path = os.path.join(test_data_path, 'pipeline1.json')

        project_id = 'project-id'
        location = 'us-central1'
        function_url = ('https://{}-{}.cloudfunctions.net/' +
                        'templated_http_request-v1').format(
                            location, project_id)

        def mock_get_cloud_functions_api():
            functions_api = mock.Mock()

            def function_get(name):
                del name
                request_mock = mock.Mock()
                request_mock.execute.return_value = {
                    'httpsTrigger': {
                        'url': function_url
                    }
                }
                return request_mock

            functions_api.get = function_get
            return functions_api

        with mock.patch(
                'kfp.v2.google.client.schedule._enable_required_apis',
                return_value=None,
        ), mock.patch(
                'kfp.v2.google.client.schedule._get_cloud_functions_api',
                new=mock_get_cloud_functions_api,
        ), mock.patch(
                'kfp.v2.google.client.schedule._create_scheduler_job',
                spec=True) as create_scheduler_job_mock:
            schedule.create_from_pipeline_file(
                pipeline_path=pipeline_path,
                schedule='46 * * * *',
                project_id=project_id,
                region=location,
                time_zone='America/Los_Angeles',
            )
            create_scheduler_job_mock.assert_called_once()
            actual_job_body = create_scheduler_job_mock.call_args[1]['job_body']
            self.assertEqual(actual_job_body['http_target']['uri'],
                             function_url)


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# `kfp` is a namespace package.
# https://packaging.python.org/guides/packaging-namespace-packages/#pkgutil-style-namespace-packages
__path__ = __import__("pkgutil").extend_path(__path__, __name__)

__version__ = '1.8.3'

from . import components
from . import containers
from . import dsl
from . import auth
from ._client import Client
from ._config import *
from ._local_client import LocalClient
from ._runners import *



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pipeline task class and operations."""

from typing import Any, Mapping, Optional

from kfp.dsl import _component_bridge
from kfp.v2.components.experimental import component_spec as cspec


# TODO(chensun): return PipelineTask object instead of ContainerOp object.
def create_pipeline_task(
    component_spec: cspec.ComponentSpec,
    arguments: Mapping[str, Any],
) -> "ContainerOp":  # pytype: disable=name-error
    return _component_bridge._create_container_op_from_component_and_arguments(
        component_spec=component_spec.to_v1_component_spec(),
        arguments=arguments,
    )


class PipelineTask:
    """Represents a pipeline task -- an instantiated component.

    Replaces `ContainerOp`. Holds operations available on a task object, such as
    `.after()`, `.set_memory_limit()`, `enable_caching()`, etc.

    Attributes:
      task_spec:
      component_spec:
      executor_spec:
    """

    def __init__(
        self,
        component_spec: cspec.ComponentSpec,
        arguments: Mapping[str, Any],
    ):
        # TODO(chensun): move logic from _component_bridge over here.
        raise NotImplementedError



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for component I/O type mapping."""
import inspect
from typing import Dict, List, Optional, Type, Union
from kfp.components import structures
from kfp.components import type_annotation_utils
from kfp.pipeline_spec import pipeline_spec_pb2
from kfp.v2.components.types import artifact_types


PARAMETER_TYPES = Union[str, int, float, bool, dict, list]

# ComponentSpec I/O types to DSL ontology artifact classes mapping.
_ARTIFACT_CLASSES_MAPPING = {
    'model': artifact_types.Model,
    'dataset': artifact_types.Dataset,
    'metrics': artifact_types.Metrics,
    'classificationmetrics': artifact_types.ClassificationMetrics,
    'slicedclassificationmetrics': artifact_types.SlicedClassificationMetrics,
    'html': artifact_types.HTML,
    'markdown': artifact_types.Markdown,
}

# ComponentSpec I/O types to (IR) PipelineTaskSpec I/O types mapping.
# The keys are normalized (lowercased). These are types viewed as Parameters.
# The values are the corresponding IR parameter primitive types.
_PARAMETER_TYPES_MAPPING = {
    'integer': pipeline_spec_pb2.PrimitiveType.INT,
    'int': pipeline_spec_pb2.PrimitiveType.INT,
    'double': pipeline_spec_pb2.PrimitiveType.DOUBLE,
    'float': pipeline_spec_pb2.PrimitiveType.DOUBLE,
    'string': pipeline_spec_pb2.PrimitiveType.STRING,
    'str': pipeline_spec_pb2.PrimitiveType.STRING,
    'text': pipeline_spec_pb2.PrimitiveType.STRING,
    'bool': pipeline_spec_pb2.PrimitiveType.STRING,
    'boolean': pipeline_spec_pb2.PrimitiveType.STRING,
    'dict': pipeline_spec_pb2.PrimitiveType.STRING,
    'list': pipeline_spec_pb2.PrimitiveType.STRING,
    'jsonobject': pipeline_spec_pb2.PrimitiveType.STRING,
    'jsonarray': pipeline_spec_pb2.PrimitiveType.STRING,
}

# Mapping primitive types to their IR message field names.
# This is used in constructing condition strings.
_PARAMETER_TYPES_VALUE_REFERENCE_MAPPING = {
    pipeline_spec_pb2.PrimitiveType.INT: 'int_value',
    pipeline_spec_pb2.PrimitiveType.DOUBLE: 'double_value',
    pipeline_spec_pb2.PrimitiveType.STRING: 'string_value',
}


def is_parameter_type(type_name: Optional[Union[str, dict]]) -> bool:
    """Check if a ComponentSpec I/O type is considered as a parameter type.

    Args:
      type_name: type name of the ComponentSpec I/O type.

    Returns:
      True if the type name maps to a parameter type else False.
    """
    if isinstance(type_name, str):
        type_name = type_annotation_utils.get_short_type_name(type_name)
    elif isinstance(type_name, dict):
        type_name = list(type_name.keys())[0]
    else:
        return False

    return type_name.lower() in _PARAMETER_TYPES_MAPPING


def get_artifact_type_schema(
    artifact_class_or_type_name: Optional[Union[str,
                                                Type[artifact_types.Artifact]]]
) -> pipeline_spec_pb2.ArtifactTypeSchema:
    """Gets the IR I/O artifact type msg for the given ComponentSpec I/O
    type."""
    artifact_class = artifact_types.Artifact
    if isinstance(artifact_class_or_type_name, str):
        artifact_class = _ARTIFACT_CLASSES_MAPPING.get(
            artifact_class_or_type_name.lower(), artifact_types.Artifact)
    elif inspect.isclass(artifact_class_or_type_name) and issubclass(
            artifact_class_or_type_name, artifact_types.Artifact):
        artifact_class = artifact_class_or_type_name

    return pipeline_spec_pb2.ArtifactTypeSchema(
        schema_title=artifact_class.TYPE_NAME,
        schema_version=artifact_class.VERSION)


def get_parameter_type(
    param_type: Optional[Union[Type, str, dict]]
) -> pipeline_spec_pb2.PrimitiveType:
    """Get the IR I/O parameter type for the given ComponentSpec I/O type.

    Args:
      param_type: type of the ComponentSpec I/O type. Can be a primitive Python
        builtin type or a type name.

    Returns:
      The enum value of the mapped IR I/O primitive type.

    Raises:
      AttributeError: if type_name is not a string type.
    """
    if type(param_type) == type:
        type_name = param_type.__name__
    elif isinstance(param_type, dict):
        type_name = list(param_type.keys())[0]
    else:
        type_name = type_annotation_utils.get_short_type_name(str(param_type))
    return _PARAMETER_TYPES_MAPPING.get(type_name.lower())


def get_parameter_type_field_name(type_name: Optional[str]) -> str:
    """Get the IR field name for the given primitive type.

    For example: 'str' -> 'string_value', 'double' -> 'double_value', etc.

    Args:
      type_name: type name of the ComponentSpec I/O primitive type.

    Returns:
      The IR value reference field name.

    Raises:
      AttributeError: if type_name is not a string type.
    """
    return _PARAMETER_TYPES_VALUE_REFERENCE_MAPPING.get(
        get_parameter_type(type_name))


def get_input_artifact_type_schema(
    input_name: str,
    inputs: List[structures.InputSpec],
) -> Optional[str]:
    """Find the input artifact type by input name.

    Args:
      input_name: The name of the component input.
      inputs: The list of InputSpec

    Returns:
      The artifact type schema of the input.

    Raises:
      AssertionError if input not found, or input found but not an artifact type.
    """
    for component_input in inputs:
        if component_input.name == input_name:
            assert not is_parameter_type(
                component_input.type), 'Input is not an artifact type.'
            return get_artifact_type_schema(component_input.type)
    assert False, 'Input not found.'



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Deprecated. See kfp.v2.components.types.type_utils instead.

This module will be removed in KFP v2.0.
"""
import warnings
from kfp.v2.components.types import type_utils

warnings.warn(
    'Module kfp.dsl.type_utils is deprecated and will be removed'
    ' in KFP v2.0. Please use from kfp.v2.components.types.type_utils instead.',
    category=FutureWarning)

is_parameter_type = type_utils.is_parameter_type
get_artifact_type_schema = type_utils.type_utils.get_artifact_type_schema
get_parameter_type = type_utils.get_parameter_type
get_parameter_type_field_name = type_utils.get_parameter_type_field_name
get_input_artifact_type_schema = type_utils.type_utils.get_input_artifact_type_schema



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import shutil
import tempfile
import unittest

from kfp import components
from kfp.v2 import compiler
from kfp.v2 import dsl
from kfp.dsl import types


VALID_PRODUCER_COMPONENT_SAMPLE = components.load_component_from_text("""
    name: producer
    inputs:
    - {name: input_param, type: String}
    outputs:
    - {name: output_model, type: Model}
    - {name: output_value, type: Integer}
    implementation:
      container:
        image: gcr.io/my-project/my-image:tag
        args:
        - {inputValue: input_param}
        - {outputPath: output_model}
        - {outputPath: output_value}
    """)

class CompilerTest(unittest.TestCase):

    def test_compile_simple_pipeline(self):

        tmpdir = tempfile.mkdtemp()
        try:
            producer_op = components.load_component_from_text("""
      name: producer
      inputs:
      - {name: input_param, type: String}
      outputs:
      - {name: output_model, type: Model}
      - {name: output_value, type: Integer}
      implementation:
        container:
          image: gcr.io/my-project/my-image:tag
          args:
          - {inputValue: input_param}
          - {outputPath: output_model}
          - {outputPath: output_value}
      """)

            consumer_op = components.load_component_from_text("""
      name: consumer
      inputs:
      - {name: input_model, type: Model}
      - {name: input_value, type: Integer}
      implementation:
        container:
          image: gcr.io/my-project/my-image:tag
          args:
          - {inputPath: input_model}
          - {inputValue: input_value}
      """)

            @dsl.pipeline(name='test-pipeline', pipeline_root='dummy_root')
            def simple_pipeline(pipeline_input: str = 'Hello KFP!'):
                producer = producer_op(input_param=pipeline_input)
                consumer = consumer_op(
                    input_model=producer.outputs['output_model'],
                    input_value=producer.outputs['output_value'])

            target_json_file = os.path.join(tmpdir, 'result.json')
            compiler.Compiler().compile(
                pipeline_func=simple_pipeline, package_path=target_json_file)

            self.assertTrue(os.path.exists(target_json_file))
        finally:
            shutil.rmtree(tmpdir)

    def test_compile_pipeline_with_dsl_graph_component_should_raise_error(self):

        with self.assertRaisesRegex(
                NotImplementedError,
                'dsl.graph_component is not yet supported in KFP v2 compiler.'):

            @dsl.component
            def flip_coin_op() -> str:
                import random
                result = 'heads' if random.randint(0, 1) == 0 else 'tails'
                return result

            @dsl.graph_component
            def flip_coin_graph_component():
                flip = flip_coin_op()
                with dsl.Condition(flip.output == 'heads'):
                    flip_coin_graph_component()

            @dsl.pipeline(name='test-pipeline', pipeline_root='dummy_root')
            def my_pipeline():
                flip_coin_graph_component()

            compiler.Compiler().compile(
                pipeline_func=my_pipeline, package_path='output.json')

    def test_compile_pipeline_with_misused_inputvalue_should_raise_error(self):

        upstream_op = components.load_component_from_text("""
        name: upstream compoent
        outputs:
        - {name: model, type: Model}
        implementation:
          container:
            image: dummy
            args:
            - {outputPath: model}
        """)
        downstream_op = components.load_component_from_text("""
        name: compoent with misused placeholder
        inputs:
        - {name: model, type: Model}
        implementation:
          container:
            image: dummy
            args:
            - {inputValue: model}
        """)

        @dsl.pipeline(name='test-pipeline', pipeline_root='dummy_root')
        def my_pipeline():
            downstream_op(model=upstream_op().output)

        with self.assertRaisesRegex(
                TypeError,
                ' type "Model" cannot be paired with InputValuePlaceholder.'):
            compiler.Compiler().compile(
                pipeline_func=my_pipeline, package_path='output.json')

    def test_compile_pipeline_with_misused_inputpath_should_raise_error(self):

        component_op = components.load_component_from_text("""
        name: compoent with misused placeholder
        inputs:
        - {name: text, type: String}
        implementation:
          container:
            image: dummy
            args:
            - {inputPath: text}
        """)

        @dsl.pipeline(name='test-pipeline', pipeline_root='dummy_root')
        def my_pipeline(text: str):
            component_op(text=text)

        with self.assertRaisesRegex(
                TypeError,
                ' type "String" cannot be paired with InputPathPlaceholder.'):
            compiler.Compiler().compile(
                pipeline_func=my_pipeline, package_path='output.json')

    def test_compile_pipeline_with_missing_task_should_raise_error(self):

        @dsl.pipeline(name='test-pipeline', pipeline_root='dummy_root')
        def my_pipeline(text: str):
            pass

        with self.assertRaisesRegex(
                ValueError,'Task is missing from pipeline.'):
            compiler.Compiler().compile(
                  pipeline_func=my_pipeline, package_path='output.json')

    def test_compile_pipeline_with_misused_inputuri_should_raise_error(self):

        component_op = components.load_component_from_text("""
        name: compoent with misused placeholder
        inputs:
        - {name: value, type: Float}
        implementation:
          container:
            image: dummy
            args:
            - {inputUri: value}
        """)

        @dsl.pipeline(name='test-pipeline', pipeline_root='dummy_root')
        def my_pipeline(value: float):
            component_op(value=value)

        with self.assertRaisesRegex(
                TypeError,
                ' type "Float" cannot be paired with InputUriPlaceholder.'):
            compiler.Compiler().compile(
                pipeline_func=my_pipeline, package_path='output.json')

    def test_compile_pipeline_with_misused_outputuri_should_raise_error(self):

        component_op = components.load_component_from_text("""
        name: compoent with misused placeholder
        outputs:
        - {name: value, type: Integer}
        implementation:
          container:
            image: dummy
            args:
            - {outputUri: value}
        """)

        @dsl.pipeline(name='test-pipeline', pipeline_root='dummy_root')
        def my_pipeline():
            component_op()

        with self.assertRaisesRegex(
                TypeError,
                ' type "Integer" cannot be paired with OutputUriPlaceholder.'):
            compiler.Compiler().compile(
                pipeline_func=my_pipeline, package_path='output.json')

    def test_compile_pipeline_with_invalid_name_should_raise_error(self):

        def my_pipeline():
            VALID_PRODUCER_COMPONENT_SAMPLE(input_param='input')

        with self.assertRaisesRegex(
                ValueError,
                'Invalid pipeline name: .*\nPlease specify a pipeline name that matches'
        ):
            compiler.Compiler().compile(
                pipeline_func=my_pipeline, package_path='output.json')

    def test_set_pipeline_root_through_pipeline_decorator(self):

        tmpdir = tempfile.mkdtemp()
        try:

            @dsl.pipeline(name='test-pipeline', pipeline_root='gs://path')
            def my_pipeline():
                VALID_PRODUCER_COMPONENT_SAMPLE(input_param='input')

            target_json_file = os.path.join(tmpdir, 'result.json')
            compiler.Compiler().compile(
                pipeline_func=my_pipeline, package_path=target_json_file)

            self.assertTrue(os.path.exists(target_json_file))
            with open(target_json_file) as f:
                job_spec = json.load(f)
            self.assertEqual('gs://path',
                             job_spec['runtimeConfig']['gcsOutputDirectory'])
        finally:
            shutil.rmtree(tmpdir)

    def test_passing_string_parameter_to_artifact_should_error(self):

        component_op = components.load_component_from_text("""
      name: compoent
      inputs:
      - {name: some_input, type: , description: an uptyped input}
      implementation:
        container:
          image: dummy
          args:
          - {inputPath: some_input}
      """)

        @dsl.pipeline(name='test-pipeline', pipeline_root='gs://path')
        def my_pipeline(input1: str):
            component_op(some_input=input1)

        with self.assertRaisesRegex(
                TypeError,
                'Passing PipelineParam "input1" with type "String" \(as "Parameter"\) '
                'to component input "some_input" with type "None" \(as "Artifact"\) is '
                'incompatible. Please fix the type of the component input.'):
            compiler.Compiler().compile(
                pipeline_func=my_pipeline, package_path='output.json')

    def test_passing_missing_type_annotation_on_pipeline_input_should_error(
            self):

        @dsl.pipeline(name='test-pipeline', pipeline_root='gs://path')
        def my_pipeline(input1):
            pass

        with self.assertRaisesRegex(
                TypeError,
                'The pipeline argument \"input1\" is viewed as an artifact due '
                'to its type \"None\". And we currently do not support passing '
                'artifacts as pipeline inputs. Consider type annotating the '
                'argument with a primitive type, such as \"str\", \"int\", '
                '\"float\", \"bool\", \"dict\", and \"list\".'):
            compiler.Compiler().compile(
                pipeline_func=my_pipeline, package_path='output.json')

    def test_passing_generic_artifact_to_input_expecting_concrete_artifact(
            self):

        producer_op1 = components.load_component_from_text("""
      name: producer compoent
      outputs:
      - {name: output, type: Artifact}
      implementation:
        container:
          image: dummy
          args:
          - {outputPath: output}
      """)

        @dsl.component
        def producer_op2(output: dsl.Output[dsl.Artifact]):
            pass

        consumer_op1 = components.load_component_from_text("""
      name: consumer compoent
      inputs:
      - {name: input, type: MyDataset}
      implementation:
        container:
          image: dummy
          args:
          - {inputPath: input}
      """)

        @dsl.component
        def consumer_op2(input: dsl.Input[dsl.Dataset]):
            pass

        @dsl.pipeline(name='test-pipeline')
        def my_pipeline():
            consumer_op1(producer_op1().output)
            consumer_op1(producer_op2().output)
            consumer_op2(producer_op1().output)
            consumer_op2(producer_op2().output)

        try:
            tmpdir = tempfile.mkdtemp()
            target_json_file = os.path.join(tmpdir, 'result.json')
            compiler.Compiler().compile(
                pipeline_func=my_pipeline, package_path=target_json_file)

            self.assertTrue(os.path.exists(target_json_file))
        finally:
            shutil.rmtree(tmpdir)

    def test_passing_arbitrary_artifact_to_input_expecting_concrete_artifact(
            self):

        producer_op1 = components.load_component_from_text("""
      name: producer compoent
      outputs:
      - {name: output, type: SomeArbitraryType}
      implementation:
        container:
          image: dummy
          args:
          - {outputPath: output}
      """)

        @dsl.component
        def consumer_op(input: dsl.Input[dsl.Dataset]):
            pass

        @dsl.pipeline(name='test-pipeline')
        def my_pipeline():
            consumer_op(producer_op1().output)
            consumer_op(producer_op2().output)

        with self.assertRaisesRegex(
                types.InconsistentTypeException,
                'Incompatible argument passed to the input "input" of component '
                '"Consumer op": Argument type "SomeArbitraryType" is incompatible with '
                'the input type "Dataset"'):
            compiler.Compiler().compile(
                pipeline_func=my_pipeline, package_path='result.json')

    def test_constructing_container_op_directly_should_error(self):

        @dsl.pipeline(name='test-pipeline')
        def my_pipeline():
            dsl.ContainerOp(
                name='comp1',
                image='gcr.io/dummy',
                command=['python', 'main.py'])

        with self.assertRaisesRegex(
                RuntimeError,
                'Constructing ContainerOp instances directly is deprecated and not '
                'supported when compiling to v2 \(using v2 compiler or v1 compiler '
                'with V2_COMPATIBLE or V2_ENGINE mode\).'):
            compiler.Compiler().compile(
                pipeline_func=my_pipeline, package_path='result.json')


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import itertools
import re
import textwrap
from typing import Callable, Dict, List, Mapping, Optional, TypeVar
import warnings

import docstring_parser

from kfp import components as v1_components
from kfp.components import _components, _data_passing, structures, type_annotation_utils
from kfp.v2.components.types import artifact_types, type_annotations

_DEFAULT_BASE_IMAGE = 'python:3.7'


def _python_function_name_to_component_name(name):
    name_with_spaces = re.sub(' +', ' ', name.replace('_', ' ')).strip(' ')
    return name_with_spaces[0].upper() + name_with_spaces[1:]


def _get_packages_to_install_command(
        package_list: Optional[List[str]] = None) -> List[str]:
    result = []
    if package_list:
        install_pip_command = 'python3 -m ensurepip'
        install_packages_command = (
            'PIP_DISABLE_PIP_VERSION_CHECK=1 python3 -m pip install --quiet \
                --no-warn-script-location {}').format(' '.join(
                [repr(str(package)) for package in package_list]))
        result = [
            'sh', '-c', '({install_pip} || {install_pip} --user) &&'
            ' ({install_packages} || {install_packages} --user) && "$0" "$@"'
            .format(
                install_pip=install_pip_command,
                install_packages=install_packages_command)
        ]
    return result


def _get_default_kfp_package_path() -> str:
    import kfp
    return 'kfp=={}'.format(kfp.__version__)


def _get_function_source_definition(func: Callable) -> str:
    func_code = inspect.getsource(func)

    # Function might be defined in some indented scope (e.g. in another
    # function). We need to handle this and properly dedent the function source
    # code
    func_code = textwrap.dedent(func_code)
    func_code_lines = func_code.split('\n')

    # Removing possible decorators (can be multiline) until the function
    # definition is found
    func_code_lines = itertools.dropwhile(lambda x: not x.startswith('def'),
                                          func_code_lines)

    if not func_code_lines:
        raise ValueError(
            'Failed to dedent and clean up the source of function "{}". '
            'It is probably not properly indented.'.format(func.__name__))

    return '\n'.join(func_code_lines)


def _annotation_to_type_struct(annotation):
    if not annotation or annotation == inspect.Parameter.empty:
        return None
    if hasattr(annotation, 'to_dict'):
        annotation = annotation.to_dict()
    if isinstance(annotation, dict):
        return annotation
    if isinstance(annotation, type):
        type_struct = _data_passing.get_canonical_type_name_for_type(annotation)
        if type_struct:
            return type_struct
        type_name = str(annotation.__name__)
    elif hasattr(
            annotation, '__forward_arg__'
    ):  # Handling typing.ForwardRef('Type_name') (the name was _ForwardRef in python 3.5-3.6)
        type_name = str(annotation.__forward_arg__)
    else:
        type_name = str(annotation)

    # It's also possible to get the converter by type name
    type_struct = _data_passing.get_canonical_type_name_for_type(type_name)
    if type_struct:
        return type_struct
    return type_name


def _maybe_make_unique(name: str, names: List[str]):
    if name not in names:
        return name

    for i in range(2, 100):
        unique_name = '{}_{}'.format(name, i)
        if unique_name not in names:
            return unique_name

    raise RuntimeError('Too many arguments with the name {}'.format(name))


# TODO(KFPv2): Replace with v2 ComponentSpec.
def _func_to_component_spec(
        func: Callable,
        base_image: Optional[str] = None,
        packages_to_install: Optional[List[str]] = None,
        install_kfp_package: bool = True,
        kfp_package_path: Optional[str] = None) -> structures.ComponentSpec:
    decorator_base_image = getattr(func, '_component_base_image', None)
    if decorator_base_image is not None:
        if base_image is not None and decorator_base_image != base_image:
            raise ValueError(
                'base_image ({}) conflicts with the decorator-specified base image metadata ({})'
                .format(base_image, decorator_base_image))
        else:
            base_image = decorator_base_image
    else:
        if base_image is None:
            base_image = _DEFAULT_BASE_IMAGE
            if isinstance(base_image, Callable):
                base_image = base_image()

    imports_source = [
        "from kfp.v2.dsl import *",
        "from typing import *",
    ]

    func_source = _get_function_source_definition(func)

    source = textwrap.dedent("""
        {imports_source}

        {func_source}\n""").format(
        imports_source='\n'.join(imports_source), func_source=func_source)

    packages_to_install = packages_to_install or []
    if install_kfp_package:
        if kfp_package_path is None:
            kfp_package_path = _get_default_kfp_package_path()
        packages_to_install.append(kfp_package_path)

    packages_to_install_command = _get_packages_to_install_command(
        package_list=packages_to_install)

    from kfp.components._structures import ExecutorInputPlaceholder
    component_spec = extract_component_interface(func)

    component_spec.implementation = structures.ContainerImplementation(
        container=structures.ContainerSpec(
            image=base_image,
            command=packages_to_install_command + [
                'sh',
                '-ec',
                textwrap.dedent('''\
                    program_path=$(mktemp -d)
                    printf "%s" "$0" > "$program_path/ephemeral_component.py"
                    python3 -m kfp.v2.components.executor_main \
                        --component_module_path \
                        "$program_path/ephemeral_component.py" \
                        "$@"
                '''),
                source,
            ],
            args=[
                "--executor_input",
                ExecutorInputPlaceholder(),
                "--function_to_execute",
                func.__name__,
            ]))
    return component_spec


def extract_component_interface(func: Callable) -> structures.ComponentSpec:
    single_output_name_const = 'Output'

    signature = inspect.signature(func)
    parameters = list(signature.parameters.values())

    parsed_docstring = docstring_parser.parse(inspect.getdoc(func))
    doc_dict = {p.arg_name: p.description for p in parsed_docstring.params}

    inputs = []
    outputs = []

    input_names = set()
    output_names = set()
    for parameter in parameters:
        parameter_type = type_annotation_utils.maybe_strip_optional_from_annotation(
            parameter.annotation)
        passing_style = None
        io_name = parameter.name

        if type_annotations.is_artifact_annotation(parameter_type):
            # passing_style is either type_annotations.InputAnnotation or
            # type_annotations.OutputAnnotation.
            passing_style = type_annotations.get_io_artifact_annotation(
                parameter_type)

            # parameter_type is type_annotations.Artifact or one of its subclasses.
            parameter_type = type_annotations.get_io_artifact_class(
                parameter_type)
            if not issubclass(parameter_type, artifact_types.Artifact):
                raise ValueError(
                    'Input[T] and Output[T] are only supported when T is a '
                    'subclass of Artifact. Found `{} with type {}`'.format(
                        io_name, parameter_type))

            if parameter.default is not inspect.Parameter.empty:
                raise ValueError(
                    'Default values for Input/Output artifacts are not supported.'
                )
        elif isinstance(parameter_type,
                        (v1_components.InputPath, v1_components.OutputPath)):
            raise TypeError(
                'In v2 components, please import the Python function'
                ' annotations `InputPath` and `OutputPath` from'
                ' package `kfp.v2.dsl` instead of `kfp.dsl`.')
        elif isinstance(
                parameter_type,
            (type_annotations.InputPath, type_annotations.OutputPath)):
            passing_style = type(parameter_type)
            parameter_type = parameter_type.type
            if parameter.default is not inspect.Parameter.empty and not (
                    passing_style == type_annotations.InputPath and
                    parameter.default is None):
                raise ValueError(
                    'Path inputs only support default values of None. Default values for outputs are not supported.'
                )

        type_struct = _annotation_to_type_struct(parameter_type)

        if passing_style in [
                type_annotations.OutputAnnotation, type_annotations.OutputPath
        ]:
            io_name = _maybe_make_unique(io_name, output_names)
            output_names.add(io_name)
            output_spec = structures.OutputSpec(
                name=io_name,
                type=type_struct,
                description=doc_dict.get(parameter.name))
            output_spec._passing_style = passing_style
            output_spec._parameter_name = parameter.name
            outputs.append(output_spec)
        else:
            io_name = _maybe_make_unique(io_name, input_names)
            input_names.add(io_name)
            input_spec = structures.InputSpec(
                name=io_name,
                type=type_struct,
                description=doc_dict.get(parameter.name))
            if parameter.default is not inspect.Parameter.empty:
                input_spec.optional = True
                if parameter.default is not None:
                    outer_type_name = list(type_struct.keys())[0] if isinstance(
                        type_struct, dict) else type_struct
                    try:
                        input_spec.default = _data_passing.serialize_value(
                            parameter.default, outer_type_name)
                    except Exception as ex:
                        warnings.warn(
                            'Could not serialize the default value of the parameter "{}". {}'
                            .format(parameter.name, ex))
            input_spec._passing_style = passing_style
            input_spec._parameter_name = parameter.name
            inputs.append(input_spec)

    #Analyzing the return type annotations.
    return_ann = signature.return_annotation
    if hasattr(return_ann, '_fields'):  #NamedTuple
        # Getting field type annotations.
        # __annotations__ does not exist in python 3.5 and earlier
        # _field_types does not exist in python 3.9 and later
        field_annotations = getattr(return_ann,
                                    '__annotations__', None) or getattr(
                                        return_ann, '_field_types', None)
        for field_name in return_ann._fields:
            type_struct = None
            if field_annotations:
                type_struct = _annotation_to_type_struct(
                    field_annotations.get(field_name, None))

            output_name = _maybe_make_unique(field_name, output_names)
            output_names.add(output_name)
            output_spec = structures.OutputSpec(
                name=output_name,
                type=type_struct,
            )
            output_spec._passing_style = None
            output_spec._return_tuple_field_name = field_name
            outputs.append(output_spec)
    # Deprecated dict-based way of declaring multiple outputs. Was only used by the @component decorator
    elif isinstance(return_ann, dict):
        warnings.warn(
            "The ability to specify multiple outputs using the dict syntax has been deprecated."
            "It will be removed soon after release 0.1.32."
            "Please use typing.NamedTuple to declare multiple outputs.")
        for output_name, output_type_annotation in return_ann.items():
            output_type_struct = _annotation_to_type_struct(
                output_type_annotation)
            output_spec = structures.OutputSpec(
                name=output_name,
                type=output_type_struct,
            )
            outputs.append(output_spec)
    elif signature.return_annotation is not None and signature.return_annotation != inspect.Parameter.empty:
        output_name = _maybe_make_unique(single_output_name_const, output_names)
        # Fixes exotic, but possible collision: `def func(output_path: OutputPath()) -> str: ...`
        output_names.add(output_name)
        type_struct = _annotation_to_type_struct(signature.return_annotation)
        output_spec = structures.OutputSpec(
            name=output_name,
            type=type_struct,
        )
        output_spec._passing_style = None
        outputs.append(output_spec)

    # Component name and description are derived from the function's name and docstring.
    # The name can be overridden by setting setting func.__name__ attribute (of the legacy func._component_human_name attribute).
    # The description can be overridden by setting the func.__doc__ attribute (or the legacy func._component_description attribute).
    component_name = getattr(func, '_component_human_name',
                             None) or _python_function_name_to_component_name(
                                 func.__name__)
    description = getattr(func, '_component_description',
                          None) or parsed_docstring.short_description
    if description:
        description = description.strip()

    component_spec = structures.ComponentSpec(
        name=component_name,
        description=description,
        inputs=inputs if inputs else None,
        outputs=outputs if outputs else None,
    )
    return component_spec


def create_component_from_func(func: Callable,
                               base_image: Optional[str] = None,
                               packages_to_install: List[str] = None,
                               output_component_file: Optional[str] = None,
                               install_kfp_package: bool = True,
                               kfp_package_path: Optional[str] = None):
    """Converts a Python function to a v2 lightweight component.

    A lightweight component is a self-contained Python function that includes
    all necessary imports and dependencies.

    Args:
        func: The python function to create a component from. The function
            should have type annotations for all its arguments, indicating how
            it is intended to be used (e.g. as an input/output Artifact object,
            a plain parameter, or a path to a file).
        base_image: The image to use when executing |func|. It should
            contain a default Python interpreter that is compatible with KFP.
        packages_to_install: A list of optional packages to install before
            executing |func|.
        install_kfp_package: Specifies if we should add a KFP Python package to
            |packages_to_install|. Lightweight Python functions always require
            an installation of KFP in |base_image| to work. If you specify
            a |base_image| that already contains KFP, you can set this to False.
        kfp_package_path: Specifies the location from which to install KFP. By
            default, this will try to install from PyPi using the same version
            as that used when this component was created. KFP developers can
            choose to override this to point to a Github pull request or
            other pip-compatible location when testing changes to lightweight
            Python functions.

    Returns:
        A component task factory that can be used in pipeline definitions.
    """
    component_spec = _func_to_component_spec(
        func=func,
        base_image=base_image,
        packages_to_install=packages_to_install,
        install_kfp_package=install_kfp_package,
        kfp_package_path=kfp_package_path)
    if output_component_file:
        component_spec.save(output_component_file)

    # TODO(KFPv2): Replace with v2 BaseComponent.
    return _components._create_task_factory_from_component_spec(component_spec)



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for AIPlatformPipelines API client."""

import datetime
import json
import pkg_resources
import re
import subprocess
import warnings
from typing import Any, Dict, List, Mapping, Optional

from absl import logging
from google import auth
from google.auth import exceptions
from google.oauth2 import credentials
from google.protobuf import json_format
from googleapiclient import discovery

from kfp.v2.google.client import client_utils
from kfp.v2.google.client import runtime_config_builder
from kfp.v2.google.client.schedule import _create_from_pipeline_dict

# AIPlatformPipelines API endpoint.
DEFAULT_ENDPOINT_FORMAT = '{region}-aiplatform.googleapis.com'

# AIPlatformPipelines API version.
DEFAULT_API_VERSION = 'v1beta1'

# If application default credential does not exist, we fall back to whatever
# previously provided to `gcloud auth`. One can run `gcloud auth login` to
# provide identity for this token.
_AUTH_ARGS = ('gcloud', 'auth', 'print-access-token')

# AIPlatformPipelines service API parent pattern.
_PARENT_PATTERN = 'projects/{}/locations/{}'

# AIPlatformPipelines service API job name relative name prefix pattern.
_JOB_NAME_PATTERN = '{parent}/pipelineJobs/{job_id}'

# Pattern for valid names used as a uCAIP resource name.
_VALID_NAME_PATTERN = re.compile('^[a-z][-a-z0-9]{0,127}$')

# Cloud Console UI link of a pipeline run.
UI_PIPELINE_JOB_LINK_FORMAT = (
    'https://console.cloud.google.com/vertex-ai/locations/{region}/pipelines/'
    'runs/{job_id}?project={project_id}')

# Display UI link HTML snippet
UI_LINK_HTML_FORMAT = (
    'See the Pipeline job <a href="{}" target="_blank" >here</a>.')


class _AccessTokenCredentials(credentials.Credentials):
    """Credential class to provide token-based authentication."""

    def __init__(self, token):
        super().__init__(token=token)

    def refresh(self, request):
        # Overrides refresh method in the base class because by default token is
        # not refreshable.
        pass


def _get_gcp_credential() -> Optional[credentials.Credentials]:
    """Returns the GCP OAuth2 credential.

    Returns:
      The credential. By default the function returns the current Application
      Default Credentials, which is located at $GOOGLE_APPLICATION_CREDENTIALS. If
      not set, this function returns the current login credential, whose token is
      created by running 'gcloud auth login'.
      For more information, see
      https://cloud.google.com/sdk/gcloud/reference/auth/application-default/print-access-token
    """
    result = None
    try:
        result, _ = auth.default()
    except exceptions.DefaultCredentialsError as e:
        logging.warning(
            'Failed to get GCP access token for application '
            'default credential: (%s). Using end user credential.', e)
        try:
            token = subprocess.check_output(_AUTH_ARGS).rstrip().decode('utf-8')
            result = _AccessTokenCredentials(token=token)
        except subprocess.CalledProcessError as e:
            # _AccessTokenCredentials won't throw CalledProcessError, so this
            # exception implies that the subprocess call is failed.
            logging.warning('Failed to get GCP access token: %s', e)
    else:
        return result


def _get_current_time() -> datetime.datetime:
    """Gets the current timestamp."""
    return datetime.datetime.now()


def _is_ipython() -> bool:
    """Returns whether we are running in notebook."""
    try:
        import IPython  # pylint: disable=g-import-not-at-top
        ipy = IPython.get_ipython()
        if ipy is None:
            return False
    except ImportError:
        return False

    return True


def _extract_job_id(job_name: str) -> Optional[str]:
    """Extracts job id from job name.

    Args:
     job_name: The full job name.

    Returns:
     The job id or None if no match found.
    """
    p = re.compile(
        'projects/(?P<project_id>.*)/locations/(?P<region>.*)/pipelineJobs/(?P<job_id>.*)'
    )
    result = p.search(job_name)
    return result.group('job_id') if result else None


def _set_enable_caching_value(pipeline_spec: Dict[str, Any],
                              enable_caching: bool) -> None:
    """Sets pipeline tasks caching options.

    Args:
     pipeline_spec: The dictionary of pipeline spec.
     enable_caching: Whether to enable caching.
    """
    for component in [pipeline_spec['root']] + list(
            pipeline_spec['components'].values()):
        if 'dag' in component:
            for task in component['dag']['tasks'].values():
                task['cachingOptions'] = {'enableCache': enable_caching}


class AIPlatformClient(object):
    """AIPlatformPipelines Unified API Client."""

    def __init__(
        self,
        project_id: str,
        region: str,
    ):
        """Constructs an AIPlatformPipelines API client.

        Args:
          project_id: GCP project ID.
          region: GCP project region.
        """
        warnings.warn(
            'AIPlatformClient will be deprecated in v1.9. Please use PipelineJob'
            ' https://googleapis.dev/python/aiplatform/latest/_modules/google/cloud/aiplatform/pipeline_jobs.html'
            ' in Vertex SDK. Install the SDK using "pip install google-cloud-aiplatform"',
            category=FutureWarning,
        )

        if not project_id:
            raise ValueError(
                'A valid GCP project ID is required to run a pipeline.')
        if not region:
            raise ValueError(
                'A valid GCP region is required to run a pipeline.')

        self._endpoint = DEFAULT_ENDPOINT_FORMAT.format(region=region)
        self._api_version = DEFAULT_API_VERSION

        self._project_id = project_id
        self._region = region
        self._parent = _PARENT_PATTERN.format(project_id, region)

        discovery_doc_path = pkg_resources.resource_filename(
            'kfp.v2.google.client',
            'discovery/aiplatform_public_google_rest_v1beta1.json')

        with open(discovery_doc_path) as f:
            discovery_doc = f.read()

        self._api_client = discovery.build_from_document(
            service=discovery_doc,
            http=None,
            credentials=_get_gcp_credential(),
        )
        self._api_client._baseUrl = 'https://{}'.format(self._endpoint)  # pylint:disable=protected-access

    def _display_job_link(self, job_id: str):
        """Display an link to UI."""
        url = UI_PIPELINE_JOB_LINK_FORMAT.format(
            region=self._region, job_id=job_id, project_id=self._project_id)
        if _is_ipython():
            import IPython  # pylint: disable=g-import-not-at-top
            html = UI_LINK_HTML_FORMAT.format(url)
            IPython.display.display(IPython.display.HTML(html))
        else:
            print('See the Pipeline job here:', url)

    def _submit_job(
        self,
        job_spec: Mapping[str, Any],
        job_id: Optional[str] = None,
    ) -> dict:
        """Submits a pipeline job to run on AIPlatformPipelines service.

        Args:
          job_spec: AIPlatformPipelines pipelineJob spec.
          job_id: Optional user-specified ID of this pipelineJob. If not provided,
            pipeline service will automatically generate a random numeric ID.

        Returns:
          The service returned PipelineJob instance.

        Raises:
          RuntimeError: If AIPlatformPipelines service returns unexpected response
          or empty job name.
        """
        request = self._api_client.projects().locations().pipelineJobs().create(
            body=job_spec, pipelineJobId=job_id, parent=self._parent)
        response = request.execute()
        job_id = _extract_job_id(response['name'])
        if job_id:
            self._display_job_link(job_id)
        return response

    def get_job(self, job_id: str) -> dict:
        """Gets an existing pipeline job on AIPlatformPipelines service.

        Args:
          job_id: The relative ID of this pipelineJob. The full qualified name will
            generated according to the project ID and region specified for the
            client.

        Returns:
          The JSON-formatted response from service representing the pipeline job.
        """
        full_name = _JOB_NAME_PATTERN.format(parent=self._parent, job_id=job_id)
        return self._api_client.projects().locations().pipelineJobs().get(
            name=full_name).execute()

    def list_jobs(self) -> dict:
        """Lists existing pipeline jobs on AIPlatformPipelines service.

        Returns:
          The JSON-formatted response from service representing all pipeline jobs.
        """
        return self._api_client.projects().locations().pipelineJobs().list(
            parent=self._parent).execute()

    def create_run_from_job_spec(
            self,
            job_spec_path: str,
            job_id: Optional[str] = None,
            pipeline_root: Optional[str] = None,
            parameter_values: Optional[Mapping[str, Any]] = None,
            enable_caching: Optional[bool] = None,
            cmek: Optional[str] = None,
            service_account: Optional[str] = None,
            network: Optional[str] = None,
            labels: Optional[Mapping[str, str]] = None) -> dict:
        """Runs a pre-compiled pipeline job on AIPlatformPipelines service.

        Args:
          job_spec_path: The path of PipelineJob JSON file. It can be a local path
            or a GS URI.
          job_id: Optionally, the user can provide the unique ID of the job run. If
            not specified, pipeline name + timestamp will be used.
          pipeline_root: Optionally the user can override the pipeline root
            specified during the compile time.
          parameter_values: The mapping from runtime parameter names to its values.
          enable_caching: Whether or not to enable caching for the run.
            If not set, defaults to the compile time settings, which are True for all
            tasks by default, while users may specify different caching options for
            individual tasks.
            If set, the setting applies to all tasks in the pipeline -- overrides
            the compile time settings.
          cmek: The customer-managed encryption key for a pipelineJob. If set, the
            pipeline job and all of its sub-resources will be secured by this key.
          service_account: The service account that the pipeline workload runs as.
          network: The network configuration applied for pipeline jobs. If left
            unspecified, the workload is not peered with any network.
          labels: The user defined metadata to organize PipelineJob.

        Returns:
          Full AIPlatformPipelines job name.

        Raises:
          ParseError: On JSON parsing problems.
          RuntimeError: If AIPlatformPipelines service returns unexpected response
          or empty job name.
        """
        job_spec = client_utils.load_json(job_spec_path)
        pipeline_name = job_spec['pipelineSpec']['pipelineInfo']['name']
        job_id = job_id or '{pipeline_name}-{timestamp}'.format(
            pipeline_name=re.sub('[^-0-9a-z]+', '-',
                                 pipeline_name.lower()).lstrip('-').rstrip('-'),
            timestamp=_get_current_time().strftime('%Y%m%d%H%M%S'))
        if not _VALID_NAME_PATTERN.match(job_id):
            raise ValueError(
                'Generated job ID: {} is illegal as a uCAIP pipelines job ID. '
                'Expecting an ID following the regex pattern '
                '"[a-z][-a-z0-9]{{0,127}}"'.format(job_id))

        job_name = _JOB_NAME_PATTERN.format(parent=self._parent, job_id=job_id)

        job_spec['name'] = job_name
        job_spec['displayName'] = job_id

        builder = runtime_config_builder.RuntimeConfigBuilder.from_job_spec_json(
            job_spec)
        builder.update_pipeline_root(pipeline_root)
        builder.update_runtime_parameters(parameter_values)

        runtime_config = builder.build()
        job_spec['runtimeConfig'] = runtime_config

        if enable_caching is not None:
            _set_enable_caching_value(job_spec['pipelineSpec'], enable_caching)

        if cmek is not None:
            job_spec['encryptionSpec'] = {'kmsKeyName': cmek}
        if service_account is not None:
            job_spec['serviceAccount'] = service_account
        if network is not None:
            job_spec['network'] = network

        if labels:
            if not isinstance(labels, Mapping):
                raise ValueError(
                    'Expect labels to be a mapping of string key value pairs. '
                    'Got "{}" of type "{}"'.format(labels, type(labels)))
            for k, v in labels.items():
                if not isinstance(k, str) or not isinstance(v, str):
                    raise ValueError(
                        'Expect labels to be a mapping of string key value pairs. '
                        'Got "{}".'.format(labels))

            job_spec['labels'] = labels

        return self._submit_job(
            job_spec=job_spec,
            job_id=job_id,
        )

    def create_schedule_from_job_spec(
        self,
        job_spec_path: str,
        schedule: str,
        time_zone: str = 'US/Pacific',
        pipeline_root: Optional[str] = None,
        parameter_values: Optional[Mapping[str, Any]] = None,
        service_account: Optional[str] = None,
        enable_caching: Optional[bool] = None,
        app_engine_region: Optional[str] = None,
        cloud_scheduler_service_account: Optional[str] = None,
    ) -> dict:
        """Creates schedule for compiled pipeline file.

        This function creates scheduled job which will run the provided pipeline on
        schedule. This is implemented by creating a Google Cloud Scheduler Job.
        The job will be visible in https://console.google.com/cloudscheduler and can
        be paused/resumed and deleted.

        To make the system work, this function also creates a Google Cloud Function
        which acts as an intermediary between the Scheduler and Pipelines. A single
        function is shared between all scheduled jobs.
        The following APIs will be activated automatically:
        * cloudfunctions.googleapis.com
        * cloudscheduler.googleapis.com
        * appengine.googleapis.com

        Args:
          job_spec_path: Path of the compiled pipeline file.
          schedule: Schedule in cron format. Example: "45 * * * *"
          time_zone: Schedule time zone. Default is 'US/Pacific'
          parameter_values: Arguments for the pipeline parameters
          pipeline_root: Optionally the user can override the pipeline root
            specified during the compile time.
          service_account: The service account that the pipeline workload runs as.
          enable_caching: Whether or not to enable caching for the run.
            If not set, defaults to the compile time settings, which are True for all
            tasks by default, while users may specify different caching options for
            individual tasks.
            If set, the setting applies to all tasks in the pipeline -- overrides
            the compile time settings.
          app_engine_region: The region that cloud scheduler job is created in.
          cloud_scheduler_service_account: The service account that Cloud Scheduler job and the proxy cloud function use.
            this should have permission to call AI Platform API and the proxy function.
            If not specified, the functions uses the App Engine default service account.

        Returns:
          Created Google Cloud Scheduler Job object dictionary.
        """
        job_spec = client_utils.load_json(job_spec_path)

        if enable_caching is not None:
            _set_enable_caching_value(job_spec['pipelineSpec'], enable_caching)

        return _create_from_pipeline_dict(
            pipeline_dict=job_spec,
            schedule=schedule,
            project_id=self._project_id,
            region=self._region,
            time_zone=time_zone,
            parameter_values=parameter_values,
            pipeline_root=pipeline_root,
            service_account=service_account,
            app_engine_region=app_engine_region,
            cloud_scheduler_service_account=cloud_scheduler_service_account)



# Copyright 2018 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re

from setuptools import setup

NAME = 'kfp'
#VERSION = .... Change the version in kfp/__init__.py

# NOTICE, after any updates to the following, ./requirements.in should be updated
# accordingly.
REQUIRES = [
    'absl-py>=0.9,<=0.11',
    'PyYAML>=5.3,<6',
    # `Blob.from_string` was introduced in google-cloud-storage 1.20.0
    # https://github.com/googleapis/python-storage/blob/master/CHANGELOG.md#1200
    'google-cloud-storage>=1.20.0,<2',
    'kubernetes>=8.0.0,<19',
    # google-api-python-client v2 doesn't work for private dicovery by default:
    # https://github.com/googleapis/google-api-python-client/issues/1225#issuecomment-791058235
    'google-api-python-client>=1.7.8,<2',
    'google-auth>=1.6.1,<2',
    'requests-toolbelt>=0.8.0,<1',
    'cloudpickle>=1.3.0,<2',
    # Update the upper version whenever a new major version of the
    # kfp-server-api package is released.
    # Update the lower version when kfp sdk depends on new apis/fields in
    # kfp-server-api.
    # Note, please also update ./requirements.in
    'kfp-server-api>=1.1.2,<2.0.0',
    'jsonschema>=3.0.1,<4',
    'tabulate>=0.8.6,<1',
    'click>=7.1.1,<8',
    'Deprecated>=1.2.7,<2',
    'strip-hints>=0.1.8,<1',
    'docstring-parser>=0.7.3,<1',
    'kfp-pipeline-spec>=0.1.10,<0.2.0',
    'fire>=0.3.1,<1',
    'protobuf>=3.13.0,<4',
    'uritemplate>=3.0.1,<4',
    # Standard library backports
    'dataclasses;python_version<"3.7"',
    'typing-extensions>=3.10.0.2,<4;python_version<"3.9"',
    'pydantic>=1.8.2,<2',
]

TESTS_REQUIRE = [
    'frozendict',
]


def find_version(*file_path_parts):
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, *file_path_parts), 'r') as fp:
        version_file_text = fp.read()

    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file_text,
        re.M,
    )
    if version_match:
        return version_match.group(1)

    raise RuntimeError('Unable to find version string.')


setup(
    name=NAME,
    version=find_version('kfp', '__init__.py'),
    description='KubeFlow Pipelines SDK',
    author='The Kubeflow Authors',
    url="https://github.com/kubeflow/pipelines",
    project_urls={
        "Documentation": "https://kubeflow-pipelines.readthedocs.io/en/stable/",
        "Bug Tracker": "https://github.com/kubeflow/pipelines/issues",
        "Source": "https://github.com/kubeflow/pipelines/tree/master/sdk",
        "Changelog": "https://github.com/kubeflow/pipelines/blob/master/sdk/RELEASE.md",
    },
    install_requires=REQUIRES,
    tests_require=TESTS_REQUIRE,
    packages=[
        'kfp',
        'kfp.auth',
        'kfp.cli',
        'kfp.cli.diagnose_me',
        'kfp.compiler',
        'kfp.components',
        'kfp.components.structures',
        'kfp.containers',
        'kfp.dsl',
        'kfp.dsl.extensions',
        'kfp.notebook',
        'kfp.v2',
        'kfp.v2.compiler',
        'kfp.v2.components',
        'kfp.v2.components.types',
        'kfp.v2.components.experimental',
        'kfp.v2.dsl',
        'kfp.v2.google.client',
        'kfp.v2.google.experimental',
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.6.1',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'dsl-compile = kfp.compiler.main:main',
            'dsl-compile-v2 = kfp.v2.compiler.main:main',
            'kfp=kfp.__main__:main'
        ]
    })



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.v2.google.experimental.custom_job."""

import unittest

from kfp.dsl import _container_op
from kfp.v2.google.experimental import run_as_aiplatform_custom_job


class CustomJobTest(unittest.TestCase):

    def test_run_as_aiplatform_custom_job_simple_mode(self):
        task = _container_op.ContainerOp(
            name='test-task',
            image='python:3.7',
            command=['python3', 'main.py'],
            arguments=['arg1', 'arg2'])
        run_as_aiplatform_custom_job(
            task,
            display_name='custom-job1',
            replica_count=10,
            machine_type='n1-standard-8',
            accelerator_type='NVIDIA_TESLA_K80',
            accelerator_count=2,
            boot_disk_type='pd-ssd',
            boot_disk_size_gb=200,
            timeout='3600s',
            restart_job_on_worker_restart=True,
            service_account='test-sa',
            network='projects/123/global/networks/mypvc',
            output_uri_prefix='gs://bucket/')

        expected_custom_job_spec = {
            'displayName': 'custom-job1',
            'jobSpec': {
                'workerPoolSpecs': [{
                    'replicaCount': '1',
                    'machineSpec': {
                        'machineType': 'n1-standard-8',
                        'acceleratorType': 'NVIDIA_TESLA_K80',
                        'acceleratorCount': 2
                    },
                    'containerSpec': {
                        'imageUri': 'python:3.7',
                        'command': ['python3', 'main.py'],
                        'args': ['arg1', 'arg2']
                    },
                    'diskSpec': {
                        'bootDiskType': 'pd-ssd',
                        'bootDiskSizeGb': 200
                    }
                }, {
                    'replicaCount': '9',
                    'machineSpec': {
                        'machineType': 'n1-standard-8',
                        'acceleratorType': 'NVIDIA_TESLA_K80',
                        'acceleratorCount': 2
                    },
                    'containerSpec': {
                        'imageUri': 'python:3.7',
                        'command': ['python3', 'main.py'],
                        'args': ['arg1', 'arg2']
                    },
                    'diskSpec': {
                        'bootDiskType': 'pd-ssd',
                        'bootDiskSizeGb': 200
                    }
                }],
                'scheduling': {
                    'timeout': '3600s',
                    'restartJobOnWorkerRestart': True
                },
                'serviceAccount': 'test-sa',
                'network': 'projects/123/global/networks/mypvc',
                'baseOutputDirectory': {
                    'outputUriPrefix': 'gs://bucket/'
                }
            }
        }
        self.maxDiff = None
        self.assertDictEqual(task.custom_job_spec, expected_custom_job_spec)

    def test_run_as_aiplatform_custom_job_use_specified_worker_pool_specs(self):
        task = _container_op.ContainerOp(
            name='test-task',
            image='python:3.7',
            command=['python3', 'main.py'],
            arguments=['arg1', 'arg2'])
        run_as_aiplatform_custom_job(
            task,
            display_name='custom-job1',
            worker_pool_specs=[
                {
                    'containerSpec': {
                        'imageUri': 'alpine',
                        'command': ['sh', '-c', 'echo 1'],
                    },
                    'replicaCount': '1',
                    'machineSpec': {
                        'machineType': 'n1-standard-8',
                    },
                },
            ])

        expected_custom_job_spec = {
            'displayName': 'custom-job1',
            'jobSpec': {
                'workerPoolSpecs': [{
                    'containerSpec': {
                        'imageUri': 'alpine',
                        'command': ['sh', '-c', 'echo 1']
                    },
                    'replicaCount': '1',
                    'machineSpec': {
                        'machineType': 'n1-standard-8'
                    }
                }]
            }
        }

        print(task.custom_job_spec)
        self.maxDiff = None
        self.assertDictEqual(task.custom_job_spec, expected_custom_job_spec)


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import types
from typing import Any, Callable, Mapping, Sequence
from inspect import Parameter, Signature


class KwParameter(Parameter):

    def __init__(self,
                 name,
                 default=Parameter.empty,
                 annotation=Parameter.empty):
        super().__init__(
            name,
            Parameter.POSITIONAL_OR_KEYWORD,
            default=default,
            annotation=annotation)


def create_function_from_parameter_names(func: Callable[[Mapping[str, Any]],
                                                        Any],
                                         parameter_names: Sequence[str],
                                         documentation=None,
                                         func_name=None,
                                         func_filename=None):
    return create_function_from_parameters(
        func, [KwParameter(name) for name in parameter_names], documentation,
        func_name, func_filename)


def create_function_from_parameters(func: Callable[[Mapping[str, Any]], Any],
                                    parameters: Sequence[Parameter],
                                    documentation=None,
                                    func_name=None,
                                    func_filename=None):
    new_signature = Signature(parameters)  # Checks the parameter consistency

    def pass_locals():
        return dict_func(locals())  # noqa: F821 TODO

    code = pass_locals.__code__
    mod_co_argcount = len(parameters)
    mod_co_nlocals = len(parameters)
    mod_co_varnames = tuple(param.name for param in parameters)
    mod_co_name = func_name or code.co_name
    if func_filename:
        mod_co_filename = func_filename
        mod_co_firstlineno = 1
    else:
        mod_co_filename = code.co_filename
        mod_co_firstlineno = code.co_firstlineno

    if sys.version_info >= (3, 8):
        modified_code = code.replace(
            co_argcount=mod_co_argcount,
            co_nlocals=mod_co_nlocals,
            co_varnames=mod_co_varnames,
            co_filename=mod_co_filename,
            co_name=mod_co_name,
            co_firstlineno=mod_co_firstlineno,
        )
    else:
        modified_code = types.CodeType(
            mod_co_argcount, code.co_kwonlyargcount, mod_co_nlocals,
            code.co_stacksize, code.co_flags, code.co_code, code.co_consts,
            code.co_names, mod_co_varnames, mod_co_filename, mod_co_name,
            mod_co_firstlineno, code.co_lnotab)

    default_arg_values = tuple(
        p.default for p in parameters if p.default != Parameter.empty
    )  #!argdefs "starts from the right"/"is right-aligned"
    modified_func = types.FunctionType(
        modified_code, {
            'dict_func': func,
            'locals': locals
        },
        name=func_name,
        argdefs=default_arg_values)
    modified_func.__doc__ = documentation
    modified_func.__signature__ = new_signature

    return modified_func



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import json
import importlib
import os
import sys

from kfp.v2.components import executor as component_executor


def _load_module(module_name: str, module_directory: str):
    """Dynamically imports the Python module with the given name and package
    path.

    E.g., Assuming there is a file called `my_module.py` under
    `/some/directory/my_module`, we can use::

        _load_module('my_module', '/some/directory')

    to effectively `import mymodule`.

    Args:
        module_name: The name of the module.
        package_path: The package under which the specified module resides.
    """
    module_spec = importlib.util.spec_from_file_location(
        name=module_name,
        location=os.path.join(module_directory, f'{module_name}.py'))
    module = importlib.util.module_from_spec(module_spec)
    sys.modules[module_spec.name] = module
    module_spec.loader.exec_module(module)
    return module


def executor_main():
    parser = argparse.ArgumentParser(description='KFP Component Executor.')
    parser.add_argument(
        '--component_module_path',
        type=str,
        help='Path to a module containing the KFP component.')
    parser.add_argument(
        '--function_to_execute',
        type=str,
        help='The name of the component function in '
        '--component_module_path file that is to be executed.')
    parser.add_argument(
        '--executor_input',
        type=str,
        help='JSON-serialized ExecutorInput from the orchestrator. '
        'This should contain inputs and placeholders for outputs.')

    args, _ = parser.parse_known_args()

    module_directory = os.path.dirname(args.component_module_path)
    module_name = os.path.basename(args.component_module_path)[:-len('.py')]
    print('Loading KFP component module {} from dir {}'.format(
        module_name, module_directory))

    module = _load_module(
        module_name=module_name, module_directory=module_directory)

    executor_input = json.loads(args.executor_input)
    function_to_execute = getattr(module, args.function_to_execute)

    executor = component_executor.Executor(
        executor_input=executor_input, function_to_execute=function_to_execute)

    executor.execute()


if __name__ == '__main__':
    executor_main()
